"""
init
"""

from pandas.plotting._core import PlotAccessor
from pandas.core.frame import DataFrame
from typing import Any
from pandas.core.dtypes.generic import ABCDataFrame
import importlib
import types

__version__ = "0.1.2"


class PlotAccessor:
    """
    Make plots of MassSpec data using dataframes

    """

    _common_kinds = ("line", "vline", "scatter")
    _msdata_kinds = ("chromatogram", "mobilogram", "spectrum", "peakmap")
    _all_kinds = _common_kinds + _msdata_kinds

    def __init__(self, data: DataFrame) -> None:
        self._parent = data

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        backend_name = kwargs.get("backend", None)
        if backend_name is None:
            backend_name = "matplotlib"

        plot_backend = _get_plot_backend(backend_name)

        x, y, kind, kwargs = self._get_call_args(
            plot_backend.__name__, self._parent, args, kwargs
        )

        if kind not in self._all_kinds:
            raise ValueError(
                f"{kind} is not a valid plot kind "
                f"Valid plot kinds: {self._all_kinds}"
            )

        # Call the plot method of the selected backend
        if "backend" in kwargs:
            kwargs.pop("backend")
        return plot_backend.plot(self._parent, x=x, y=y, kind=kind, **kwargs)

    @staticmethod
    def _get_call_args(backend_name: str, data: DataFrame, args, kwargs):
        """
        Get the arguments to pass to the plotting backend.

        Parameters
        ----------
        backend_name : str
            The name of the backend.
        data : DataFrame
            The data to plot.
        args : tuple
            The positional arguments passed to the plotting function.
        kwargs : dict
            The keyword arguments passed to the plotting function.

        Returns
        -------
        dict
            The arguments to pass to the plotting backend.
        """
        if isinstance(data, ABCDataFrame):
            arg_def = [
                ("x", None),
                ("y", None),
                ("kind", "line"),
                ("by", None),
                ("subplots", None),
                ("sharex", None),
                ("sharey", None),
                ("height", None),
                ("width", None),
                ("grid", None),
                ("toolbar_location", None),
                ("fig", None),
                ("title", None),
                ("xlabel", None),
                ("ylabel", None),
                ("x_axis_location", None),
                ("y_axis_location", None),
                ("line_type", None),
                ("line_width", None),
                ("min_border", None),
                ("show_plot", None),
                ("legend", None),
                ("feature_config", None),
                ("_config", None),
                ("backend", backend_name),
            ]
        else:
            raise ValueError(
                f"Unsupported data type: {type(data).__name__}, expected DataFrame."
            )

        pos_args = {name: value for (name, _), value in zip(arg_def, args)}

        kwargs = dict(arg_def, **pos_args, **kwargs)

        x = kwargs.pop("x", None)
        y = kwargs.pop("y", None)
        kind = kwargs.pop("kind", "line")
        return x, y, kind, kwargs


_backends: dict[str, types.ModuleType] = {}


def _load_backend(backend: str) -> types.ModuleType:
    """
    Load a plotting backend.

    Parameters
    ----------
    backend : str
        The identifier for the backend. Either "bokeh", "matplotlib", "plotly",
        or a module name.

    Returns
    -------
    types.ModuleType
        The imported backend.
    """
    if backend == "bokeh":
        try:
            module = importlib.import_module("pyopenms_viz._bokeh")
        except ImportError:
            raise ImportError(
                "Bokeh is required for plotting when the 'bokeh' backend is selected."
            ) from None
        return module

    elif backend == "matplotlib":
        try:
            module = importlib.import_module("pyopenms_viz._matplotlib")
        except ImportError:
            raise ImportError(
                "Matplotlib is required for plotting when the 'matplotlib' backend is selected."
            ) from None
        return module

    elif backend == "plotly":
        try:
            module = importlib.import_module("pyopenms_viz._plotly")
        except ImportError:
            raise ImportError(
                "Plotly is required for plotting when the 'plotly' backend is selected."
            ) from None
        return module

    raise ValueError(
        f"Could not find plotting backend '{backend}'. Needs to be one of 'bokeh', 'matplotlib', or 'plotly'."
    )


def _get_plot_backend(backend: str | None = None):

    backend_str: str = backend or "matplotlib"

    if backend_str in _backends:
        return _backends[backend_str]

    module = _load_backend(backend_str)
    _backends[backend_str] = module
    return module


__all__ = ["PlotAccessor"]



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Tuple, Literal, Dict, Any
from enum import Enum
from copy import deepcopy


# TODO: This class could be removed, or changed to backend and then the config can be interpolated into the dataframe plot call
class Engine(Enum):
    PLOTLY = 1
    BOKEH = 2
    MATPLOTLIB = 3


def bokeh_line_dash_mapper(bokeh_dash, target_library="plotly"):
    """
    Maps Bokeh line dash types to their Plotly or Matplotlib equivalents.

    Args:
    bokeh_dash (str or list): Bokeh line dash type or custom dash pattern.
    target_library (str): 'plotly' or 'matplotlib' (default: 'plotly')

    Returns:
    str or list or tuple: Equivalent line dash type for the target library.
    """
    plotly_mapper = {
        "solid": "solid",
        "dashed": "dash",
        "dotted": "dot",
        "dotdash": "dashdot",
        "dashdot": "dashdot",
    }

    matplotlib_mapper = {
        "solid": "-",
        "dashed": "--",
        "dotted": ":",
        "dotdash": "-.",
        "dashdot": "-.",
    }

    if target_library.lower() == "plotly":
        if isinstance(bokeh_dash, str):
            # If it's already a valid Plotly dash type, return it as is
            if bokeh_dash in plotly_mapper.values():
                return bokeh_dash
            # Otherwise, map from Bokeh to Plotly
            return plotly_mapper.get(bokeh_dash, "solid")
        elif isinstance(bokeh_dash, list):
            return " ".join(f"{num}px" for num in bokeh_dash)
    elif target_library.lower() == "matplotlib":
        if isinstance(bokeh_dash, str):
            # If it's already a valid Matplotlib dash type, return it as is
            if bokeh_dash in matplotlib_mapper.values():
                return bokeh_dash
            # Otherwise, map from Bokeh to Matplotlib
            return matplotlib_mapper.get(bokeh_dash, "-")
        elif isinstance(bokeh_dash, list):
            return (None, tuple(bokeh_dash))

    # Default return if target_library is not recognized or bokeh_dash is neither string nor list
    return "solid" if target_library.lower() == "plotly" else "-"


@dataclass(kw_only=True)
class LegendConfig:
    loc: str = "right"
    orientation: str = "vertical"
    title: str = "Legend"
    fontsize: int = 10
    show: bool = True
    onClick: Literal["hide", "mute"] = (
        "mute"  # legend click policy, only valid for bokeh
    )
    bbox_to_anchor: Tuple[float, float] = (
        1.2,
        0.5,
    )  # for fine control legend positioning in matplotlib

    @staticmethod
    def _matplotlibLegendLocationMapper(loc):
        """
        Maps the legend location to the matplotlib equivalent
        """
        loc_mapper = {
            "right": "center right",
            "left": "center left",
            "above": "upper center",
            "below": "lower center",
        }
        return loc_mapper[loc]

    @classmethod
    def from_dict(cls, legend_dict: Dict[str, Any]) -> "LegendConfig":
        """
        Convert a dictionary to a LegendConfig instance.

        Args:
        legend_dict (Dict[str, Any]): Dictionary containing legend configuration.

        Returns:
        LegendConfig: An instance of LegendConfig with the specified settings.
        """
        # Create a dictionary with default values
        config = {
            "loc": "right",
            "orientation": "vertical",
            "title": "Legend",
            "fontsize": 10,
            "show": True,
            "onClick": "mute",
            "bbox_to_anchor": (1.2, 0.5),
        }

        # Update with provided values
        config.update(legend_dict)

        # Ensure onClick is a valid value
        if config["onClick"] not in ["hide", "mute"]:
            config["onClick"] = "mute"

        # Create and return the LegendConfig instance
        return cls(**config)


@dataclass(kw_only=True)
class FeatureConfig:
    def default_legend_factory():
        return LegendConfig(title="Features", loc="right", bbox_to_anchor=(1.5, 0.5))

    colormap: str = "viridis"
    line_width: float = 1
    line_type: str = "solid"
    legend: LegendConfig = field(default_factory=default_legend_factory)

    @classmethod
    def from_dict(cls, feature_dict: Dict[str, Any]) -> "FeatureConfig":
        """
        Convert a dictionary to a FeatureConfig instance.

        Args:
        feature_dict (Dict[str, Any]): Dictionary containing feature configuration.

        Returns:
        FeatureConfig: An instance of FeatureConfig with the specified settings.
        """
        # Extract the legend dictionary if it exists
        legend_dict = feature_dict.pop("legend", None)

        # Create the LegendConfig instance if legend_dict is provided
        if legend_dict:
            legend_config = LegendConfig.from_dict(legend_dict)
        else:
            legend_config = cls.default_legend_factory()

        # Create a dictionary with default values
        config = {
            "colormap": "viridis",
            "line_width": 1,
            "line_type": "solid",
            "legend": legend_config,
        }

        # Update with provided values
        config.update(feature_dict)

        # Create and return the FeatureConfig instance
        return cls(**config)


@dataclass(kw_only=True)
class _BasePlotConfig(ABC):

    def default_legend_factory():
        return LegendConfig(title="Trace")

    kind: str = "default"
    title: str = "1D Plot"
    xlabel: str = "X-axis"
    ylabel: str = "Y-axis"
    x_axis_location: str = "below"
    y_axis_location: str = "left"
    min_border: str = 0
    engine: Literal["PLOTLY", "BOKEH", "MATPLOTLIB"] = "PLOTLY"
    height: int = 500
    width: int = 500
    relative_intensity: bool = False
    show_legend: bool = True
    show_plot: bool = True
    colormap: str = "viridis"
    grid: bool = True
    line_type: str = "solid"
    line_width: float = 1
    toolbar_location: str = "above"

    legend: LegendConfig = field(default_factory=default_legend_factory)
    feature_config: FeatureConfig = field(default_factory=FeatureConfig)

    def __post_init__(self):
        # Update default plot labels based on the kind of plot
        self.set_plot_labels()

    @property
    def engine_enum(self):
        return Engine[self.engine]

    def copy(self):
        return deepcopy(self)

    def set_plot_labels(self):
        """
        Set the title, xlabel, and ylabel based on the kind of plot.
        """
        plot_configs = {
            "chromatogram": {
                "title": "Chromatogram",
                "xlabel": "Retention Time",
                "ylabel": "Intensity",
            },
            "mobilogram": {
                "title": "Mobilogram",
                "xlabel": "Ion Mobility",
                "ylabel": "Intensity",
            },
            "spectrum": {
                "title": "Mass Spectrum",
                "xlabel": "mass-to-charge",
                "ylabel": "Intensity",
            },
            "peakmap": {
                "title": "PeakMap",
                "xlabel": "Retention Time",
                "ylabel": "mass-to-charge",
            },
            # Add more plot types as needed
        }

        if self.kind in plot_configs:
            self.title = plot_configs[self.kind]["title"]
            self.xlabel = plot_configs[self.kind]["xlabel"]
            self.ylabel = plot_configs[self.kind]["ylabel"]

            if self.relative_intensity and "Intensity" in self.ylabel:
                self.ylabel = "Relative " + self.ylabel



from __future__ import annotations

from abc import ABC

from typing import Tuple

from bokeh.plotting import figure
from bokeh.palettes import Plasma256
from bokeh.transform import linear_cmap
from bokeh.models import (
    ColumnDataSource,
    Legend,
    Range1d,
    BoxEditTool,
    Span,
    VStrip,
    GlyphRenderer,
    Label,
)

from pandas.core.frame import DataFrame

# pyopenms_viz imports
from .._core import (
    BasePlot,
    LinePlot,
    VLinePlot,
    ScatterPlot,
    BaseMSPlot,
    ChromatogramPlot,
    MobilogramPlot,
    PeakMapPlot,
    SpectrumPlot,
    APPEND_PLOT_DOC,
)
from .._misc import ColorGenerator, MarkerShapeGenerator
from ..constants import PEAK_BOUNDARY_ICON, FEATURE_BOUNDARY_ICON


class BOKEHPlot(BasePlot, ABC):
    """
    Base class for assembling a Bokeh plot
    """

    def _interactive(self):
        return False

    def _load_extension(self) -> None:
        try:
            from bokeh.plotting import figure, show
            from bokeh.models import ColumnDataSource, Legend
        except ImportError:
            raise ImportError(
                f"bokeh is not installed. Please install using `pip install bokeh` to use this plotting library in pyopenms-viz"
            )

    def _create_figure(self) -> None:
        if self.fig is None:
            self.fig = figure(
                title=self.title,
                x_axis_label=self.xlabel,
                y_axis_label=self.ylabel,
                x_axis_location=self.x_axis_location,
                y_axis_location=self.y_axis_location,
                width=self.width,
                height=self.height,
                min_border=self.min_border,
            )

    def _update_plot_aes(self, fig, **kwargs):
        """
        Update the aesthetics of the plot
        """
        fig.grid.visible = self.grid
        fig.toolbar_location = self.toolbar_location

    def _add_legend(self, fig, legend):
        """
        Add the legend
        """
        if self.legend.show:
            fig.add_layout(legend, self.legend.loc)
            fig.legend.orientation = self.legend.orientation
            fig.legend.click_policy = self.legend.onClick
            fig.legend.title = self.legend.title
            fig.legend.label_text_font_size = str(self.legend.fontsize) + "pt"

    def _add_tooltips(self, fig, tooltips, custom_hover_data=None):
        """
        Add tooltips to the plot
        """
        from bokeh.models import HoverTool

        hover = HoverTool()
        hover.tooltips = tooltips
        fig.add_tools(hover)

    def _add_bounding_box_drawer(self, fig, **kwargs):
        """
        Add a BoxEditTool to the figure for drawing bounding boxes.

        Args:
            fig (figure): The Bokeh figure object to add the BoxEditTool to.

        Returns:
            The renderer object that is used to draw the bounding box.
        """
        r = fig.rect(
            [],
            [],
            [],
            [],
            fill_alpha=0,
            line_dash="dashed",
            line_width=3,
            line_color="#F02D1A",
        )
        draw_tool = BoxEditTool(renderers=[r], empty_value=0)
        # TODO: change how icon path is defined
        draw_tool.icon = FEATURE_BOUNDARY_ICON
        draw_tool.name = "Draw Bounding Box"

        # Add the tool to the figure
        fig.add_tools(draw_tool)
        return r

    def _add_bounding_vertical_drawer(self, fig, **kwargs):
        """
        Add a BoxEditTool to the figure for drawing bounding vertical strips.

        Args:
            fig (figure): The Bokeh figure object to add the BoxEditTool to.

        Returns:
            The renderer object that is used to draw the bounding box.
        """
        # Create empty source data
        source = ColumnDataSource(data=dict(x0=[], x1=[]))

        # Create the VStrip glyph
        glyph = VStrip(
            x0="x0",
            x1="x1",
            fill_alpha=0,
            line_dash="dashed",
            line_width=3,
            line_color="#F02D1A",
        )

        # Create a GlyphRenderer for the VStrip glyph using the same data source
        renderer = GlyphRenderer(data_source=source, glyph=glyph)

        # Add the GlyphRenderer to the fig object's renderers list
        fig.renderers.append(renderer)

        draw_tool = BoxEditTool(renderers=[renderer], empty_value=0)
        # TODO: change how icon path is defined
        draw_tool.icon = PEAK_BOUNDARY_ICON
        draw_tool.name = "Draw Peak Boundary Strip"

        # Add the tool to the figure
        fig.add_tools(draw_tool)
        return renderer

    def _modify_x_range(
        self, x_range: Tuple[float, float], padding: Tuple[float, float] | None = None
    ):
        """
        Modify the x-axis range.

        Args:
            x_range (Tuple[float, float]): The desired x-axis range.
            padding (Tuple[float, float] | None, optional): The padding to be applied to the x-axis range, in decimal percent. Defaults to None.
        """
        start, end = x_range
        if padding is not None:
            start = start - (start * padding[0])
            end = end + (end * padding[1])
        self.fig.x_range = Range1d(start=start, end=end)

    def _modify_y_range(
        self, y_range: Tuple[float, float], padding: Tuple[float, float] | None = None
    ):
        """
        Modifies the y-axis range of the plot.

        Args:
            y_range (Tuple[float, float]): The desired y-axis range.
            padding (Tuple[float, float] | None, optional): The padding to be applied to the y-axis range, in decimal percent. Defaults to None.
        """
        start, end = y_range
        if padding is not None:
            start = start - (start * padding[0])
            end = end + (end * padding[1])
        self.fig.y_range = Range1d(start=start, end=end)

    def show(self):
        from bokeh.io import show

        def app(doc):
            doc.add_root(self.fig)

        show(app)


class BOKEHLinePlot(BOKEHPlot, LinePlot):
    """
    Class for assembling a collection of Bokeh line plots
    """

    @classmethod
    @APPEND_PLOT_DOC
    def plot(cls, fig, data, x, y, by: str | None = None, **kwargs):
        """
        Plot a line plot
        """
        color_gen = kwargs.pop("line_color", None)

        if by is None:
            source = ColumnDataSource(data)
            if color_gen is not None:
                kwargs["line_color"] = next(color_gen)
            line = fig.line(x=x, y=y, source=source, **kwargs)

            return fig, None
        else:

            legend_items = []
            for group, df in data.groupby(by):
                source = ColumnDataSource(df)
                if color_gen is not None:
                    kwargs["line_color"] = next(color_gen)
                line = fig.line(x=x, y=y, source=source, **kwargs)
                legend_items.append((group, [line]))

            legend = Legend(items=legend_items)

            return fig, legend


class BOKEHVLinePlot(BOKEHPlot, VLinePlot):
    """
    Class for assembling a series of vertical line plots in Bokeh
    """

    @classmethod
    @APPEND_PLOT_DOC
    def plot(cls, fig, data, x, y, by: str | None = None, **kwargs):
        """
        Plot a set of vertical lines
        """
        color_gen = kwargs.pop("line_color", None)
        if color_gen is None:
            color_gen = ColorGenerator()
        data["line_color"] = [next(color_gen) for _ in range(len(data))]
        if by is None:
            source = ColumnDataSource(data)
            line = fig.segment(
                x0=x,
                y0=0,
                x1=x,
                y1=y,
                source=source,
                line_color="line_color",
                **kwargs,
            )
            return fig, None
        else:
            legend_items = []
            for group, df in data.groupby(by):
                source = ColumnDataSource(df)
                line = fig.segment(
                    x0=x,
                    y0=0,
                    x1=x,
                    y1=y,
                    source=source,
                    line_color="line_color",
                    **kwargs,
                )
                legend_items.append((group, [line]))

            legend = Legend(items=legend_items)

            return fig, legend

    def _add_annotations(
        self,
        fig,
        ann_texts: list[str],
        ann_xs: list[float],
        ann_ys: list[float],
        ann_colors: list[str],
    ):
        for text, x, y, color in zip(ann_texts, ann_xs, ann_ys, ann_colors):
            label = Label(
                x=x,
                y=y,
                text=text,
                text_font_size="8pt",
                text_color=color,
                x_offset=1,
                y_offset=0,
            )
            fig.add_layout(label)


class BOKEHScatterPlot(BOKEHPlot, ScatterPlot):
    """
    Class for assembling a Bokeh scatter plot
    """

    @classmethod
    @APPEND_PLOT_DOC
    def plot(cls, fig, data, x, y, by: str | None = None, **kwargs):
        """
        Plot a scatter plot
        """
        z = kwargs.pop("z", None)

        color_gen = kwargs.pop("color_gen", None)
        if color_gen is None:
            color_gen = ColorGenerator()

        marker_gen = kwargs.pop("marker_gen", None)
        if marker_gen is None:
            marker_gen = MarkerShapeGenerator(engine="BOKEH")

        mapper = linear_cmap(
            field_name=z,
            palette=Plasma256[::-1],
            low=min(data[z]),
            high=max(data[z]),
        )
        # Set defaults if they have not been set in kwargs
        defaults = {"size": 10, "line_width": 0, "fill_color": mapper if z is not None else next(color_gen)}
        for k, v in defaults.items():
            if k not in kwargs.keys():
                kwargs[k] = v

        if by is None:
            kwargs["marker"] = next(marker_gen)
            source = ColumnDataSource(data)
            line = fig.scatter(x=x, y=y, source=source, **kwargs)
            return fig, None
        else:
            legend_items = []
            for group, df in data.groupby(by):
                kwargs["marker"] = next(marker_gen)
                source = ColumnDataSource(df)
                line = fig.scatter(x=x, y=y, source=source, **kwargs)
                legend_items.append((group, [line]))
            legend = Legend(items=legend_items)

            return fig, legend


class BOKEH_MSPlot(BaseMSPlot, BOKEHPlot, ABC):

    def get_line_renderer(self, data, x, y, **kwargs) -> None:
        return BOKEHLinePlot(data, x, y, **kwargs)

    def get_vline_renderer(self, data, x, y, **kwargs) -> None:
        return BOKEHVLinePlot(data, x, y, **kwargs)

    def get_scatter_renderer(self, data, x, y, **kwargs) -> None:
        return BOKEHScatterPlot(data, x, y, **kwargs)

    def plot_x_axis_line(self, fig):
        zero_line = Span(
            location=0, dimension="width", line_color="#EEEEEE", line_width=1.5
        )
        fig.add_layout(zero_line)

    def _create_tooltips(self, entries, index=True):
        # Tooltips for interactive information
        tooltips = []
        if index:
            tooltips.append(("index", "$index"))
        for key, value in entries.items():
            tooltips.append((key, f"@{value}"))
        return tooltips, None


class BOKEHChromatogramPlot(BOKEH_MSPlot, ChromatogramPlot):
    """
    Class for assembling a Bokeh extracted ion chromatogram plot
    """

    def _add_peak_boundaries(self, annotation_data):
        """
        Add peak boundaries to the plot.

        Args:
            annotation_data (DataFrame): The feature data containing the peak boundaries.

        Returns:
            None
        """
        color_gen = ColorGenerator(
            colormap=self.feature_config.colormap, n=annotation_data.shape[0]
        )
        legend_items = []
        for idx, (_, feature) in enumerate(annotation_data.iterrows()):
            peak_boundary_lines = self.fig.segment(
                x0=[feature["leftWidth"], feature["rightWidth"]],
                y0=[0, 0],
                x1=[feature["leftWidth"], feature["rightWidth"]],
                y1=[feature["apexIntensity"], feature["apexIntensity"]],
                color=next(color_gen),
                line_dash=self.feature_config.line_type,
                line_width=self.feature_config.line_width,
            )
            if "name" in annotation_data.columns:
                use_name = feature["name"]
            else:
                use_name = f"Feature {idx}"
            if "q_value" in annotation_data.columns:
                legend_label = f"{use_name} (q-value: {feature['q_value']:.4f})"
            else:
                legend_label = f"{use_name}"
            legend_items.append((legend_label, [peak_boundary_lines]))

        if self.feature_config.legend.show:
            legend = Legend(items=legend_items)
            legend.click_policy = self.feature_config.legend.onClick
            legend.title = self.feature_config.legend.title
            legend.orientation = self.feature_config.legend.orientation
            legend.label_text_font_size = (
                str(self.feature_config.legend.fontsize) + "pt"
            )
            self.fig.add_layout(legend, self.feature_config.legend.loc)

    def get_manual_bounding_box_coords(self):
        # Get the original data source
        data_source = self.manual_boundary_renderer.data_source

        # Make a copy of the data since we don't want to modify the original active documents data source
        bbox_data = data_source.data.copy()

        # Return the modified copy
        return DataFrame(bbox_data).rename(
            columns={"x0": "leftWidth", "x1": "rightWidth"}
        )


class BOKEHMobilogramPlot(BOKEHChromatogramPlot, MobilogramPlot):
    """
    Class for assembling a Bokeh mobilogram plot
    """

    pass


class BOKEHSpectrumPlot(BOKEH_MSPlot, SpectrumPlot):
    """
    Class for assembling a Bokeh spectrum plot
    """

    pass


class BOKEHPeakMapPlot(BOKEH_MSPlot, PeakMapPlot):
    """
    Class for assembling a Bokeh feature heatmap plot
    """

    def create_main_plot(self, x, y, z, class_kwargs, other_kwargs):
        scatterPlot = self.get_scatter_renderer(self.data, x, y, **class_kwargs)

        self.fig = scatterPlot.generate(z=z, **other_kwargs)

        if self.annotation_data is not None:
            self._add_box_boundaries(self.annotation_data)

        tooltips, _ = self._create_tooltips({self.xlabel: x, self.ylabel: y, "intensity": z})

        self._add_tooltips(self.fig, tooltips)

    def create_x_axis_plot(self, x, z, class_kwargs):
        x_fig = super().create_x_axis_plot(x, z, class_kwargs)

        # Modify plot
        x_fig.x_range = self.fig.x_range
        x_fig.width = self.fig.width
        x_fig.xaxis.visible = False
        x_fig.min_border = 0
        return x_fig

    def create_y_axis_plot(self, y, z, class_kwargs):
        y_fig = super().create_y_axis_plot(y, z, class_kwargs)

        # Modify plot
        y_fig.y_range = self.fig.y_range
        y_fig.height = self.fig.height
        y_fig.legend.orientation = "horizontal"
        y_fig.x_range.flipped = True
        y_fig.min_border = 0
        return y_fig

    def combine_plots(self, x_fig, y_fig):
        # Modify the main plot
        self.fig.yaxis.visible = False

        from bokeh.layouts import gridplot

        self.fig = gridplot([[None, x_fig], [y_fig, self.fig]])

    def get_scatter_renderer(self, data, x, y, **kwargs):
        return BOKEHScatterPlot(data, x, y, **kwargs)

    def get_manual_bounding_box_coords(self):
        # Get the original data source
        data_source = self.manual_bbox_renderer.data_source

        # Make a copy of the data since we don't want to modify the original active documents data source
        bbox_data = data_source.data.copy()

        x1 = []
        y1 = []
        for i in range(0, len(bbox_data["x"])):
            x1.append(bbox_data["x"][i] + bbox_data["width"][i])
            y1.append(bbox_data["y"][i] + bbox_data["height"][i])

        # Return the modified copy
        return DataFrame(
            {"x0": bbox_data["x"], "x1": x1, "y0": bbox_data["y"], "y1": y1}
        ).rename(
            columns={
                "x0": f"{self.x}_0",
                "x1": f"{self.x}_1",
                "y0": f"{self.y}_0",
                "y1": f"{self.y}_1",
            }
        )

    def _add_box_boundaries(self, annotation_data):
        color_gen = ColorGenerator(
            colormap=self.feature_config.colormap, n=annotation_data.shape[0]
        )
        legend_items = []
        for idx, (_, feature) in enumerate(annotation_data.iterrows()):
            x0 = feature["leftWidth"]
            x1 = feature["rightWidth"]
            y0 = feature["IM_leftWidth"]
            y1 = feature["IM_rightWidth"]

            # Calculate center points and dimensions
            center_x = (x0 + x1) / 2
            center_y = (y0 + y1) / 2
            width = abs(x1 - x0)
            height = abs(y1 - y0)

            box_boundary_lines = self.fig.rect(
                x=center_x,
                y=center_y,
                width=width,
                height=height,
                color=next(color_gen),
                line_dash=self.feature_config.line_type,
                line_width=self.feature_config.line_width,
                fill_alpha=0,
            )
            if "name" in annotation_data.columns:
                use_name = feature["name"]
            else:
                use_name = f"Feature {idx}"
            if "q_value" in annotation_data.columns:
                legend_label = f"{use_name} (q-value: {feature['q_value']:.4f})"
            else:
                legend_label = f"{use_name}"
            legend_items.append((legend_label, [box_boundary_lines]))

        if self.feature_config.legend.show:
            legend = Legend(items=legend_items)
            legend.click_policy = self.feature_config.legend.onClick
            legend.title = self.feature_config.legend.title
            legend.orientation = self.feature_config.legend.orientation
            legend.label_text_font_size = (
                str(self.feature_config.legend.fontsize) + "pt"
            )
            self.fig.add_layout(legend, self.feature_config.legend.loc)



from __future__ import annotations

from typing import TYPE_CHECKING
from ..constants import IS_SPHINX_BUILD

from .core import (
    PLOTLYLinePlot,
    PLOTLYVLinePlot,
    PLOTLYScatterPlot,
    PLOTLYChromatogramPlot,
    PLOTLYMobilogramPlot,
    PLOTLYSpectrumPlot,
    PLOTLYPeakMapPlot,
)

if TYPE_CHECKING:
    from .core import PLOTLYPlot

if IS_SPHINX_BUILD:
    from .core import PLOTLY_MSPlot, PLOTLYPlot

PLOT_CLASSES: dict[str, type[PLOTLYPlot]] = {
    "line": PLOTLYLinePlot,
    "vline": PLOTLYVLinePlot,
    "scatter": PLOTLYScatterPlot,
    "chromatogram": PLOTLYChromatogramPlot,
    "mobilogram": PLOTLYMobilogramPlot,
    "spectrum": PLOTLYSpectrumPlot,
    "peakmap": PLOTLYPeakMapPlot,
}


def plot(data, kind, **kwargs):
    plot_obj = PLOT_CLASSES[kind](data, **kwargs)
    return plot_obj


__all__ = ["plot"]



from __future__ import annotations


from typing import TYPE_CHECKING
from ..constants import IS_SPHINX_BUILD

from .core import (
    BOKEHLinePlot,
    BOKEHVLinePlot,
    BOKEHScatterPlot,
    BOKEHChromatogramPlot,
    BOKEHMobilogramPlot,
    BOKEHSpectrumPlot,
    BOKEHPeakMapPlot,
)

if TYPE_CHECKING:
    from .core import BOKEHPlot

if IS_SPHINX_BUILD:
    from .core import BOKEH_MSPlot, BOKEHPlot


PLOT_CLASSES: dict[str, type[BOKEHPlot]] = {
    "line": BOKEHLinePlot,
    "vline": BOKEHVLinePlot,
    "scatter": BOKEHScatterPlot,
    "chromatogram": BOKEHChromatogramPlot,
    "mobilogram": BOKEHMobilogramPlot,
    "spectrum": BOKEHSpectrumPlot,
    "peakmap": BOKEHPeakMapPlot,
}


def plot(data, kind, **kwargs):
    plot_obj = PLOT_CLASSES[kind](data, **kwargs)
    return plot_obj


__all__ = ["plot"]



"""
pyopenms_viz/constants
~~~~~~~~~~~~~~~~~~
"""

import os
from PIL import Image


PYOPENMS_VIZ_DIRNAME = os.path.dirname(__file__)

######################
## Icons
PEAK_BOUNDARY_ICON = Image.open(
    os.path.normpath(os.path.join(PYOPENMS_VIZ_DIRNAME, "assets/img/peak_boundary.png"))
)
FEATURE_BOUNDARY_ICON = Image.open(
    os.path.normpath(
        os.path.join(PYOPENMS_VIZ_DIRNAME, "assets/img/feature_boundary.png")
    )
)


######################
## Determine if running in SPHINX build
IS_SPHINX_BUILD = False
try:
    import sphinx

    IS_SPHINX_BUILD = hasattr(sphinx, "application")
except ImportError:
    pass  # Not running SPHINX



from __future__ import annotations

from typing import TYPE_CHECKING
from ..constants import IS_SPHINX_BUILD

from .core import (
    MATPLOTLIBLinePlot,
    MATPLOTLIBVLinePlot,
    MATPLOTLIBScatterPlot,
    MATPLOTLIBChromatogramPlot,
    MATPLOTLIBMobilogramPlot,
    MATPLOTLIBSpectrumPlot,
    MATPLOTLIBPeakMapPlot,
)

if TYPE_CHECKING:
    from .core import MATPLOTLIBPlot

if IS_SPHINX_BUILD:
    from .core import MATPLOTLIB_MSPlot, MATPLOTLIBPlot

PLOT_CLASSES: dict[str, type[MATPLOTLIBPlot]] = {
    "line": MATPLOTLIBLinePlot,
    "vline": MATPLOTLIBVLinePlot,
    "scatter": MATPLOTLIBScatterPlot,
    "chromatogram": MATPLOTLIBChromatogramPlot,
    "mobilogram": MATPLOTLIBMobilogramPlot,
    "spectrum": MATPLOTLIBSpectrumPlot,
    "peakmap": MATPLOTLIBPeakMapPlot,
}


def plot(data, kind, **kwargs):
    plot_obj = PLOT_CLASSES[kind](data, **kwargs)
    return plot_obj


__all__ = ["plot"]



from __future__ import annotations

from abc import ABC

from typing import List, Tuple, Union

import plotly.graph_objects as go
from plotly.graph_objs import Figure
from plotly.subplots import make_subplots

from pandas.core.frame import DataFrame

from numpy import column_stack

from .._core import (
    BasePlot,
    LinePlot,
    VLinePlot,
    ScatterPlot,
    BaseMSPlot,
    ChromatogramPlot,
    MobilogramPlot,
    SpectrumPlot,
    PeakMapPlot,
    APPEND_PLOT_DOC,
)

from .._config import bokeh_line_dash_mapper
from .._misc import ColorGenerator, MarkerShapeGenerator
from ..constants import PEAK_BOUNDARY_ICON, FEATURE_BOUNDARY_ICON


class PLOTLYPlot(BasePlot, ABC):
    """
    Base class for assembling a Ploty plot
    """

    @property
    def _interactive(self) -> bool:
        return True

    def _load_extension(self):
        """
        Tries to load the plotly extensions, if not throw an import error
        """
        try:
            import plotly.graph_objects
        except ImportError:
            raise ImportError(
                f"plotly is not installed. Please install using `pip install plotly` to use this plotting library in pyopenms-viz"
            )

    def _create_figure(self):
        """
        Create a new figure, if a figure is not supplied
        """
        if self.fig is None:
            self.fig = go.Figure()
            self.fig.update_layout(
                title=self.title,
                xaxis_title=self.xlabel,
                yaxis_title=self.ylabel,
                width=self.width,
                height=self.height,
            )

    def _update_plot_aes(self, fig, **kwargs) -> None:
        """
        Update the plot aesthetics.
        """
        fig.update_layout(
            legend_title=self.legend.title,
            legend_font_size=self.legend.fontsize,
            showlegend=self.legend.show,
        )

        # Update to look similar to Bokeh theme
        # Customize the layout
        fig.update_layout(
            plot_bgcolor="#FFFFFF",  # Set the plot background color
            font_family="Helvetica",  # Set the font family
            font_size=12,  # Set the font size
            title_font_family="Helvetica",  # Set the title font family
            title_font_size=16,  # Set the title font size
            xaxis_title_font_family="Helvetica",  # Set the x-axis title font family
            xaxis_title_font_size=14,  # Set the x-axis title font size
            yaxis_title_font_family="Helvetica",  # Set the y-axis title font family
            yaxis_title_font_size=14,  # Set the y-axis title font size
            xaxis_gridcolor="#CCCCCC",  # Set the x-axis grid color
            yaxis_gridcolor="#CCCCCC",  # Set the y-axis grid color
            xaxis_tickfont_family="Helvetica",  # Set the x-axis tick font family
            yaxis_tickfont_family="Helvetica",  # Set the y-axis tick font family
            legend_font_family="Helvetica",  # Set the legend font family
        )

        # Add x-axis grid lines and ticks
        fig.update_xaxes(
            showgrid=self.grid,  # Add x-axis grid lines
            showline=True,
            linewidth=1,
            linecolor="black",
            ticks="outside",  # Add x-axis ticks outside the plot area
            tickwidth=1,  # Set the width of x-axis ticks
            tickcolor="black",  # Set the color of x-axis ticks
        )

        # Add y-axis grid lines and ticks
        fig.update_yaxes(
            showgrid=self.grid,  # Add y-axis grid lines
            showline=True,
            linewidth=1,
            linecolor="black",
            tickwidth=1,  # Set the width of y-axis ticks
            tickcolor="black",  # Set the color of y-axis ticks
        )

    def _add_legend(self, fig, legend):
        pass

    def _add_tooltips(self, fig, tooltips, custom_hover_data=None):
        # In case figure is constructed of multiple traces (e.g. one trace per MS peak) add annotation for each point in trace
        if len(fig.data) > 1:
            for i in range(len(fig.data)):
                fig.data[i].update(
                    hovertemplate=tooltips,
                    customdata=[custom_hover_data[i, :]] * len(fig.data[i].x),
                )
            return
        fig.update_traces(hovertemplate=tooltips, customdata=custom_hover_data)

    def _add_bounding_box_drawer(self, fig, **kwargs):
        fig.update_layout(
            modebar_add=["drawrect", "eraseshape"],
            newshape=dict(
                showlegend=True,
                label=dict(
                    texttemplate="x0: %{x0:.2f} | x1: %{x1:.2f}<br>y0: %{y0:.2f} | y1: %{y1:.2f}",
                    textposition="top left",
                ),
                line_color="#F02D1A",
                fillcolor=None,
                line=dict(
                    dash="dash",
                ),
                opacity=0.5,
            ),
        )

    def _add_bounding_vertical_drawer(self, fig, **kwargs):
        # Note: self.label_suffix must be defined

        fig.add_trace(go.Scatter(x=[], y=[], mode="lines"))
        fig.update_layout(
            modebar_add=["drawrect", "eraseshape"],
            newshape=dict(
                showlegend=True,
                label=dict(
                    texttemplate=self.label_suffix
                    + "_0: %{x0:.2f} | "
                    + self.label_suffix
                    + "_1: %{x1:.2f}",
                    textposition="top left",
                ),
                line_color="#F02D1A",
                fillcolor=None,
                line=dict(
                    dash="dash",
                ),
                drawdirection="vertical",
                opacity=0.5,
            ),
        )

    def _modify_x_range(
        self,
        x_range: Tuple[float, float] | None = None,
        padding: Tuple[float, float] | None = None,
    ):
        start, end = x_range
        if padding is not None:
            start = start - (start * padding[0])
            end = end + (end * padding[1])
        self.fig.update_xaxes(range=[start, end])

    def _modify_y_range(
        self,
        y_range: Tuple[float, float] | None = None,
        padding: Tuple[float, float] | None = None,
    ):
        start, end = y_range
        if padding is not None:
            start = start - (start * padding[0])
            end = end + (end * padding[1])
        self.fig.update_yaxes(range=[start, end])

    def show(self, **kwargs):
        self.fig.show(**kwargs)


class PLOTLYLinePlot(PLOTLYPlot, LinePlot):
    """
    Class for assembling a set of line plots in plotly
    """

    @classmethod
    @APPEND_PLOT_DOC
    def plot(  # type: ignore[override]
        cls,
        fig,
        data: DataFrame,
        x: Union[str, int],
        y: Union[str, int],
        by: str | None = None,
        **kwargs,
    ) -> Tuple[Figure, "Legend"]:  # note legend is always none for consistency
        color_gen = kwargs.pop("line_color", None)

        traces = []
        if by is None:
            trace = go.Scatter(
                x=data[x], y=data[y], mode="lines", line=dict(color=next(color_gen))
            )
            traces.append(trace)
        else:
            for group, df in data.groupby(by):
                trace = go.Scatter(
                    x=df[x],
                    y=df[y],
                    mode="lines",
                    name=group,
                    line=dict(color=next(color_gen)),
                )
                traces.append(trace)

        fig.add_traces(data=traces)
        return fig, None


class PLOTLYVLinePlot(PLOTLYPlot, VLinePlot):

    @classmethod
    @APPEND_PLOT_DOC
    def plot(cls, fig, data, x, y, by=None, **kwargs) -> Tuple[Figure, "Legend"]:
        color_gen = kwargs.pop("line_color", None)
        traces = []
        if by is None:
            for _, row in data.iterrows():
                line_color = next(color_gen)
                trace = go.Scattergl(
                    x=[row[x]] * 2,
                    y=[0, row[y]],
                    mode="lines",
                    name="",
                    showlegend=False,
                    line=dict(color=line_color),
                )
                first_group_trace_showlenged = False
                traces.append(trace)
        else:
            for group, df in data.groupby(by):
                if "showlegend" in kwargs:
                    showlegend = kwargs["showlegend"]
                    first_group_trace_showlenged = showlegend
                else:
                    first_group_trace_showlenged = True
                for _, row in df.iterrows():
                    line_color = next(color_gen)
                    trace = go.Scattergl(
                        x=[row[x]] * 2,
                        y=[0, row[y]],
                        mode="lines",
                        name=group,
                        legendgroup=group,
                        showlegend=first_group_trace_showlenged,
                        line=dict(color=line_color),
                    )
                    first_group_trace_showlenged = False
                    traces.append(trace)

        fig.add_traces(data=traces)
        return fig, None

    def _add_annotations(
        self,
        fig,
        ann_texts: list[str],
        ann_xs: list[float],
        ann_ys: list[float],
        ann_colors: list[str],
    ):
        annotations = []
        for text, x, y, color in zip(ann_texts, ann_xs, ann_ys, ann_colors):
            annotation = go.layout.Annotation(
                text=text.replace("\n", "<br>"),
                x=x,
                y=y,
                showarrow=False,
                xanchor="left",
                font=dict(
                    family="Open Sans Mono, monospace",
                    size=12,
                    color=color,
                ),
            )
            annotations.append(annotation)

        for annotation in annotations:
            fig.add_annotation(annotation)


class PLOTLYScatterPlot(PLOTLYPlot, ScatterPlot):

    @classmethod
    @APPEND_PLOT_DOC
    def plot(cls, fig, data, x, y, by=None, **kwargs) -> Tuple[Figure, "Legend"]:
        color_gen = kwargs.pop("line_color", None)
        if color_gen is None:
            color_gen = ColorGenerator()
        marker_gen = kwargs.pop("marker_gen", None)
        if marker_gen is None:
            marker_gen = MarkerShapeGenerator(engine="PLOTLY")
        marker_dict = kwargs.pop("marker", dict())
        # Check for z-dimension and plot heatmap
        z = kwargs.pop("z", None)
        # Plotting heatmaps with z dimension overwrites marker_dict.
        if z:
            # Default values for heatmap
            heatmap_defaults = dict(
                color=data[z],
                colorscale="Inferno_r",
                showscale=False,
                size=10,
                opacity=0.8,
                cmin=data[z].min(),
                cmax=data[z].max(),
            )
            # If no marker_dict was in kwargs, use default for heatmpas
            if not marker_dict:
                marker_dict = heatmap_defaults
            # Else update existing marker dict with default values if key is missing
            else:
                for k, v in heatmap_defaults.items():
                    if k not in marker_dict.keys():
                        marker_dict[k] = v

        marker_dict["color"] = data[z] if z else next(color_gen)
        traces = []
        if by is None:
            marker_dict["symbol"] = next(marker_gen)
            trace = go.Scattergl(
                x=data[x], y=data[y], mode="markers", marker=marker_dict, showlegend=False
            )
            traces.append(trace)
        else:
            for group, df in data.groupby(by):
                marker_dict["symbol"] = next(marker_gen)
                marker_dict["color"] = next(color_gen)
                if z is not None:
                    marker_dict["color"] = df[z]
                trace = go.Scatter(
                    x=df[x],
                    y=df[y],
                    mode="markers",
                    name=group,
                    marker=marker_dict,
                    **kwargs,
                )
                traces.append(trace)

        fig.add_traces(data=traces)
        return fig, None


class PLOTLY_MSPlot(BaseMSPlot, PLOTLYPlot, ABC):

    def get_line_renderer(self, data, x, y, **kwargs) -> None:
        return PLOTLYLinePlot(data, x, y, **kwargs)

    def get_vline_renderer(self, data, x, y, **kwargs) -> None:
        return PLOTLYVLinePlot(data, x, y, **kwargs)

    def get_scatter_renderer(self, data, x, y, **kwargs) -> None:
        return PLOTLYScatterPlot(data, x, y, **kwargs)

    def plot_x_axis_line(self, fig):
        fig.add_hline(y=0, line_color="black", line=dict(width=1))

    def _create_tooltips(self, entries, index=True):
        custom_hover_data = []
        # Add data from index if required
        if index:
            custom_hover_data.append(self.data.index)
        # Get the rest of the columns
        custom_hover_data += [self.data[col] for col in entries.values()]

        tooltips = []
        # Add tooltip text for index if required
        if index:
            tooltips.append("index: %{customdata[0]}")

        custom_hover_data_index = 1 if index else 0

        for key in entries.keys():
            tooltips.append(
                f"{key}" + ": %{customdata[" + str(custom_hover_data_index) + "]}"
            )
            custom_hover_data_index += 1

        return "<br>".join(tooltips), column_stack(custom_hover_data)


class PLOTLYChromatogramPlot(PLOTLY_MSPlot, ChromatogramPlot):

    def _add_peak_boundaries(self, annotation_data, **kwargs):
        color_gen = ColorGenerator(
            colormap=self.feature_config.colormap, n=annotation_data.shape[0]
        )
        for idx, (_, feature) in enumerate(annotation_data.iterrows()):
            if "q_value" in annotation_data.columns:
                legend_label = f"Feature {idx} (q-value: {feature['q_value']:.4f})"
            else:
                legend_label = f"Feature {idx}"
            self.fig.add_trace(
                go.Scatter(
                    mode="lines",
                    x=[
                        feature["leftWidth"],
                        feature["leftWidth"],
                        feature["rightWidth"],
                        feature["rightWidth"],
                    ],
                    y=[feature["apexIntensity"], 0, 0, feature["apexIntensity"]],
                    opacity=0.5,
                    line=dict(
                        color=next(color_gen),
                        dash=bokeh_line_dash_mapper(
                            self.feature_config.line_type, "plotly"
                        ),
                        width=self.feature_config.line_width,
                    ),
                    name=legend_label,
                )
            )

    def get_manual_bounding_box_coords(self, arg):
        # TODO: Implement this method, plotly doesn't have a direct easy way of extracting the relayout events. Would need to implement / add a dash dependency to add a callback to extract the relayout events
        pass


class PLOTLYMobilogramPlot(PLOTLYChromatogramPlot, MobilogramPlot):
    pass


class PLOTLYSpectrumPlot(PLOTLY_MSPlot, SpectrumPlot):
    pass


class PLOTLYPeakMapPlot(PLOTLY_MSPlot, PeakMapPlot):

    def create_main_plot(self, x, y, z, class_kwargs, other_kwargs):
        scatterPlot = self.get_scatter_renderer(self.data, x, y, **class_kwargs)
        self.fig = scatterPlot.generate(z=z, **other_kwargs)

        tooltips, custom_hover_data = self._create_tooltips({self.xlabel: x, self.ylabel: y, "intensity": z})

        self._add_tooltips(self.fig, tooltips, custom_hover_data=custom_hover_data)

        if self.annotation_data is not None:
            self._add_box_boundaries(self.annotation_data)

    def create_x_axis_plot(self, x, z, class_kwargs) -> "figure":
        x_fig = super().create_x_axis_plot(x, z, class_kwargs)
        x_fig.update_xaxes(visible=False)

        return x_fig

    def create_y_axis_plot(self, y, z, class_kwargs) -> "figure":
        y_fig = super().create_y_axis_plot(y, z, class_kwargs)
        y_fig.update_xaxes(range=[0, self.data[z].max()])
        y_fig.update_yaxes(range=[self.data[y].min(), self.data[y].max()])
        y_fig.update_layout(xaxis_title=self.ylabel, yaxis_title=self.zlabel)

        return y_fig

    def combine_plots(self, x_fig, y_fig):
        #############
        ##  Combine Plots

        # Create a figure with subplots
        fig_m = make_subplots(
            rows=2,
            cols=2,
            shared_xaxes=True,
            shared_yaxes=True,
            vertical_spacing=0,
            horizontal_spacing=0,
            subplot_titles=(
                None,
                f"Integrated {self.xlabel}",
                f"Integrated {self.ylabel}",
                None,
            ),
            specs=[
                [{}, {"type": "xy", "rowspan": 1, "secondary_y": True}],
                [
                    {"type": "xy", "rowspan": 1, "secondary_y": False},
                    {"type": "xy", "rowspan": 1, "secondary_y": False},
                ],
            ],
        )

        # Add the heatmap to the first row
        for trace in self.fig.data:
            trace.showlegend = False
            trace.legendgroup = trace.name
            fig_m.add_trace(trace, row=2, col=2, secondary_y=False)

        # Update the heatmao layout
        fig_m.update_layout(self.fig.layout)
        fig_m.update_yaxes(row=2, col=2, secondary_y=False)

        # Add the x-axis plot to the second row
        for trace in x_fig.data:

            trace.legendgroup = trace.name
            fig_m.add_trace(trace, row=1, col=2, secondary_y=True)

        # Update the XIC layout
        fig_m.update_layout(x_fig.layout)

        # Make the y-axis of fig_xic independent
        fig_m.update_yaxes(overwrite=True, row=1, col=2, secondary_y=True)

        # Manually adjust the domain of secondary y-axis to only span the first row of the subplot
        fig_m["layout"]["yaxis3"]["domain"] = [0.5, 1.0]

        # Add the XIM plot to the second row
        for trace in y_fig.data:
            trace.showlegend = False
            trace.legendgroup = trace.name
            fig_m.add_trace(trace, row=2, col=1)

        # Update the XIM layout
        fig_m.update_layout(y_fig.layout)

        # Make the x-axis of fig_xim independent
        fig_m.update_xaxes(overwrite=True, row=2, col=1)

        # Reverse the x-axis range for the XIM subplot
        fig_m.update_xaxes(autorange="reversed", row=2, col=1)

        # Update xaxis properties
        fig_m.update_xaxes(title_text=self.xlabel, row=2, col=2)
        fig_m.update_xaxes(title_text=self.zlabel, row=2, col=1)

        # Update yaxis properties
        fig_m.update_yaxes(title_text=self.zlabel, row=1, col=2)
        fig_m.update_yaxes(title_text=self.ylabel, row=2, col=1)

        # Remove axes for first quadrant
        fig_m.update_xaxes(visible=False, row=1, col=1)
        fig_m.update_yaxes(visible=False, row=1, col=1)

        # Update the layout
        fig_m.update_layout(height=self.height, width=self.width, title=self.title)

        # Overwrite the figure with the new grid figure
        self.fig = fig_m

        self._update_plot_aes(self.fig)

    def _add_box_boundaries(self, annotation_data, **kwargs):
        color_gen = ColorGenerator(
            colormap=self.feature_config.colormap, n=annotation_data.shape[0]
        )
        for idx, (_, feature) in enumerate(annotation_data.iterrows()):
            x0 = feature["leftWidth"]
            x1 = feature["rightWidth"]
            y0 = feature["IM_leftWidth"]
            y1 = feature["IM_rightWidth"]

            color = next(color_gen)

            if "name" in annotation_data.columns:
                use_name = feature["name"]
            else:
                use_name = f"Feature {idx}"
            if "q_value" in annotation_data.columns:
                legend_label = f"{use_name} (q-value: {feature['q_value']:.4f})"
            else:
                legend_label = f"{use_name}"
            self.fig.add_trace(
                go.Scatter(
                    x=[
                        x0,
                        x1,
                        x1,
                        x0,
                        x0,
                    ],  # Start and end at the same point to close the shape
                    y=[y0, y0, y1, y1, y0],
                    mode="lines",
                    fill="none",
                    opacity=0.5,
                    line=dict(
                        color=color,
                        width=self.feature_config.line_width,
                        dash=bokeh_line_dash_mapper(
                            self.feature_config.line_type, "plotly"
                        ),
                    ),
                    name=legend_label,
                )
            )



from __future__ import annotations

from abc import ABC
from typing import Tuple

import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.patches import Rectangle

from .._config import LegendConfig

from .._misc import ColorGenerator, MarkerShapeGenerator
from .._core import (
    BasePlot,
    LinePlot,
    VLinePlot,
    ScatterPlot,
    BaseMSPlot,
    ChromatogramPlot,
    MobilogramPlot,
    SpectrumPlot,
    PeakMapPlot,
    APPEND_PLOT_DOC,
)


class MATPLOTLIBPlot(BasePlot, ABC):
    """
    Base class for assembling a Matplotlib plot.

    Attributes:
        data (DataFrame): The input data frame.
    """

    @property
    def _interactive(self):
        return False

    def _load_extension(self):
        """
        Load the matplotlib extension.
        """
        try:
            from matplotlib import pyplot
        except ImportError:
            raise ImportError(
                f"matplotlib is not installed. Please install using `pip install matplotlib` to use this plotting library in pyopenms-viz"
            )

    def _create_figure(self):
        """
        Create a figure and axes objects,
        for consistency with other backends, the self.fig object stores the matplotlib axes object
        """
        if self.fig is None and self.width is not None and self.height is not None:
            self.superFig, self.fig = plt.subplots(
                figsize=(self.width / 100, self.height / 100), dpi=100
            )
            self.fig.set_title(self.title)
            self.fig.set_xlabel(self.xlabel)
            self.fig.set_ylabel(self.ylabel)

    def _update_plot_aes(self, ax, **kwargs):
        """
        Update the plot aesthetics.

        Args:
            ax: The axes object.
            **kwargs: Additional keyword arguments.
        """
        ax.grid(self.grid, zorder=0)

    def _add_legend(self, ax, legend):
        """
        Add a legend to the plot.

        Args:
            ax: The axes object.
            legend: The legend configuration.
        """
        if legend is not None and self.legend.show:
            matplotlibLegendLoc = LegendConfig._matplotlibLegendLocationMapper(
                self.legend.loc
            )
            if self.legend.orientation == "horizontal":
                ncol = len(legend[0])
            else:
                ncol = 1

            legend = ax.legend(
                *legend,
                loc=matplotlibLegendLoc,
                title=self.legend.title,
                prop={"size": self.legend.fontsize},
                bbox_to_anchor=self.legend.bbox_to_anchor,
                ncol=ncol,
            )

            legend.get_title().set_fontsize(str(self.legend.fontsize))

    def _modify_x_range(
        self, x_range: Tuple[float, float], padding: Tuple[float, float] | None = None
    ):
        """
        Modify the x-axis range.

        Args:
            x_range (Tuple[float, float]): The x-axis range.
            padding (Tuple[float, float] | None, optional): The padding for the range. Defaults to None.
        """
        start, end = x_range
        if padding is not None:
            start = start - (start * padding[0])
            end = end + (end * padding[1])
        self.fig.set_xlim(start, end)

    def _modify_y_range(
        self, y_range: Tuple[float, float], padding: Tuple[float, float] | None = None
    ):
        """
        Modify the y-axis range.

        Args:
            y_range (Tuple[float, float]): The y-axis range.
            padding (Tuple[float, float] | None, optional): The padding for the range. Defaults to None.
        """
        start, end = y_range
        if padding is not None:
            start = start - (start * padding[0])
            end = end + (end * padding[1])
        self.fig.set_ylim(start, end)

    # since matplotlib creates static plots, we don't need to implement the following methods
    def _add_tooltips(self, fig, tooltips):
        raise NotImplementedError(
            "Matplotlib does not support interactive plots and cannot use method '_add_tooltips'"
        )

    def _add_bounding_box_drawer(self, fig, **kwargs):
        raise NotImplementedError(
            "Matplotlib does not support interactive plots and cannot use method '_add_bounding_box_drawer'"
        )

    def _add_bounding_vertical_drawer(self, fig, **kwargs):
        raise NotImplementedError(
            "Matplotlib does not support interactive plots and cannot use method '_add_bounding_vertical_drawer'"
        )

    def show(self):
        """
        Show the plot.
        """
        plt.show()


class MATPLOTLIBLinePlot(MATPLOTLIBPlot, LinePlot):
    """
    Class for assembling a matplotlib line plot
    """

    @classmethod
    @APPEND_PLOT_DOC
    def plot(  # type: ignore[override]
        cls, ax, data, x, y, by: str | None = None, **kwargs
    ) -> Tuple[Axes, "Legend"]:
        """
        Plot a line plot
        """

        color_gen = kwargs.pop("line_color", None)

        legend_lines = []
        legend_labels = []

        if by is None:
            (line,) = ax.plot(data[x], data[y], color=next(color_gen))

            return ax, None
        else:
            for group, df in data.groupby(by):
                (line,) = ax.plot(df[x], df[y], color=next(color_gen))
                legend_lines.append(line)
                legend_labels.append(group)
            return ax, (legend_lines, legend_labels)


class MATPLOTLIBVLinePlot(MATPLOTLIBPlot, VLinePlot):
    """
    Class for assembling a matplotlib vertical line plot
    """

    @classmethod
    @APPEND_PLOT_DOC
    def plot(
        cls, ax, data, x, y, by: str | None = None, **kwargs
    ) -> Tuple[Axes, "Legend"]:
        """
        Plot a vertical line
        """
        color_gen = kwargs.pop("line_color", None)

        legend_lines = []
        legend_labels = []

        if by is None:
            for _, row in data.iterrows():
                (line,) = ax.plot([row[x], row[x]], [0, row[y]], color=next(color_gen))

            return ax, None
        else:
            for group, df in data.groupby(by):
                for _, row in df.iterrows():
                    (line,) = ax.plot(
                        [row[x], row[x]], [0, row[y]], color=next(color_gen)
                    )
                legend_lines.append(line)
                legend_labels.append(group)
            return ax, (legend_lines, legend_labels)

    def _add_annotations(
        self,
        fig,
        ann_texts: list[list[str]],
        ann_xs: list[float],
        ann_ys: list[float],
        ann_colors: list[str],
    ):
        for text, x, y, color in zip(ann_texts, ann_xs, ann_ys, ann_colors):
            fig.annotate(
                text,
                xy=(x, y),
                xytext=(3, 0),
                textcoords="offset points",
                fontsize=8,
                color=color,
            )


class MATPLOTLIBScatterPlot(MATPLOTLIBPlot, ScatterPlot):
    """
    Class for assembling a matplotlib scatter plot
    """

    @classmethod
    @APPEND_PLOT_DOC
    def plot(
        cls, ax, data, x, y, by: str | None = None, **kwargs
    ) -> Tuple[Axes, "Legend"]:
        """
        Plot a scatter plot
        """
        # Colors
        color_gen = kwargs.pop("line_color", None)
        # Marker shapes
        shape_gen = kwargs.pop("shape_gen", None)
        if color_gen is None:
            color_gen = ColorGenerator()
        if shape_gen is None:
            shape_gen = MarkerShapeGenerator(engine="MATPLOTLIB")
        # Heatmap data and default config values
        z = kwargs.pop("z", None)
        if z is not None:
            for k, v in dict(
                # marker="s",
                s=30,
                edgecolors="none",
                cmap="magma_r",
            ).items():
                if k not in kwargs.keys():
                    kwargs[k] = v

        kwargs["zorder"] = 2

        legend_lines = []
        legend_labels = []
        if by is None:
            if "marker" not in kwargs.keys():
                kwargs["marker"] = next(shape_gen)
            if z is not None:
                use_color = data[z]
            else:
                use_color = next(color_gen)

            scatter = ax.scatter(data[x], data[y], c=use_color, **kwargs)

            return ax, None
        else:
            vmin, vmax = data[z].min(), data[z].max()
            for group, df in data.groupby(by):
                if z is not None:
                    use_color = df[z].values
                kwargs["marker"] = next(shape_gen)
                # Normalize colors if z is specified
                if z is not None:
                    normalize = plt.Normalize(vmin=vmin, vmax=vmax)
                    scatter = ax.scatter(
                        df[x],
                        df[y],
                        c=use_color,
                        norm=normalize,
                        **kwargs,
                    )
                else:
                    scatter = ax.scatter(
                        df[x], df[y], c=use_color, **kwargs
                    )
                legend_lines.append(scatter)
                legend_labels.append(group)
            return ax, (legend_lines, legend_labels)


class MATPLOTLIB_MSPlot(BaseMSPlot, MATPLOTLIBPlot, ABC):

    def get_line_renderer(self, data, x, y, **kwargs) -> None:
        return MATPLOTLIBLinePlot(data, x, y, **kwargs)

    def get_vline_renderer(self, data, x, y, **kwargs) -> None:
        return MATPLOTLIBVLinePlot(data, x, y, **kwargs)

    def get_scatter_renderer(self, data, x, y, **kwargs) -> None:
        return MATPLOTLIBScatterPlot(data, x, y, **kwargs)

    def plot_x_axis_line(self, fig):
        fig.plot(fig.get_xlim(), [0, 0], color="#EEEEEE", linewidth=1.5)

    def _create_tooltips(self, entries, index=True):
        # No tooltips for MATPLOTLIB because it is not interactive
        return None, None


@APPEND_PLOT_DOC
class MATPLOTLIBChromatogramPlot(MATPLOTLIB_MSPlot, ChromatogramPlot):
    """
    Class for assembling a matplotlib extracted ion chromatogram plot
    """

    def _add_peak_boundaries(self, annotation_data):
        """
        Add peak boundaries to the plot.

        Args:
            annotation_data (DataFrame): The feature data containing the peak boundaries.

        Returns:
            None
        """
        if self.by is not None and self.legend.show:
            legend = self.fig.get_legend()
            self.fig.add_artist(legend)

        color_gen = ColorGenerator(
            colormap=self.feature_config.colormap, n=annotation_data.shape[0]
        )

        legend_items = []
        legend_labels = []
        for idx, (_, feature) in enumerate(annotation_data.iterrows()):
            use_color = next(color_gen)
            left_vlne = self.fig.vlines(
                x=feature["leftWidth"],
                ymin=0,
                ymax=self.data[self.y].max(),
                lw=self.feature_config.line_width,
                color=use_color,
                ls=self.feature_config.line_type,
            )
            self.fig.vlines(
                x=feature["rightWidth"],
                ymin=0,
                ymax=self.data[self.y].max(),
                lw=self.feature_config.line_width,
                color=use_color,
                ls=self.feature_config.line_type,
            )
            legend_items.append(left_vlne)

            if "name" in annotation_data.columns:
                use_name = feature["name"]
            else:
                use_name = f"Feature {idx}"
            if "q_value" in annotation_data.columns:
                cur_legend_labels = f"{use_name} (q-value: {feature['q_value']:.4f})"
            else:
                cur_legend_labels = f"{use_name}"
            legend_labels.append(cur_legend_labels)

        if self.feature_config.legend.show:
            matplotlibLegendLoc = LegendConfig._matplotlibLegendLocationMapper(
                self.feature_config.legend.loc
            )
            self.fig.legend(
                legend_items,
                legend_labels,
                loc=matplotlibLegendLoc,
                title=self.feature_config.legend.title,
                prop={"size": self.feature_config.legend.fontsize},
                bbox_to_anchor=self.feature_config.legend.bbox_to_anchor,
            )

        # since matplotlib is not interactive cannot implement the following methods
        def get_manual_bounding_box_coords(self):
            pass


class MATPLOTLIBMobilogramPlot(MATPLOTLIBChromatogramPlot, MobilogramPlot):
    """
    Class for assembling a matplotlib mobilogram plot
    """

    pass


@APPEND_PLOT_DOC
class MATPLOTLIBSpectrumPlot(MATPLOTLIB_MSPlot, SpectrumPlot):
    """
    Class for assembling a matplotlib spectrum plot
    """

    pass


class MATPLOTLIBPeakMapPlot(MATPLOTLIB_MSPlot, PeakMapPlot):
    """
    Class for assembling a matplotlib feature heatmap plot
    """

    # override creating figure because create a 2 by 2 figure
    def _create_figure(self):
        # Create a 2 by 2 figure and axis for marginal plots
        if self.add_marginals:
            self.superFig, self.ax_grid = plt.subplots(
                2, 2, figsize=(self.width / 100, self.height / 100), dpi=200
            )
        else:
            super()._create_figure()

    def plot(self, x, y, z, **kwargs):
        super().plot(x, y, z, **kwargs)

        if self.add_marginals:
            self.ax_grid[0, 0].remove()
            self.ax_grid[0, 0].axis("off")
            # Update the figure size
            self.superFig.set_size_inches(self.width / 100, self.height / 100)
            self.superFig.subplots_adjust(wspace=0, hspace=0)

    def combine_plots(
        self, x_fig, y_fig
    ):  # plots all plotted on same figure do not need to combine
        pass

    def create_x_axis_plot(self, x, z, class_kwargs) -> "figure":
        class_kwargs["fig"] = self.ax_grid[0, 1]
        super().create_x_axis_plot(x, z, class_kwargs)

        self.ax_grid[0, 1].set_title(None)
        self.ax_grid[0, 1].set_xlabel(None)
        self.ax_grid[0, 1].set_xticklabels([])
        self.ax_grid[0, 1].set_xticks([])
        self.ax_grid[0, 1].set_ylabel(self.zlabel)
        self.ax_grid[0, 1].yaxis.set_ticks_position("right")
        self.ax_grid[0, 1].yaxis.set_label_position("right")
        self.ax_grid[0, 1].yaxis.tick_right()
        self.ax_grid[0, 1].legend_ = None

    def create_y_axis_plot(self, y, z, class_kwargs) -> "figure":
        # Note y_config is different so we cannot use the base class methods
        class_kwargs["fig"] = self.ax_grid[1, 0]
        group_cols = [y]
        if self.by is not None:
            group_cols.append(self.by)

        y_data = self._integrate_data_along_dim(self.data, group_cols, z)
        y_config = self._config.copy()
        y_config.xlabel = self.zlabel
        y_config.ylabel = self.ylabel
        y_config.y_axis_location = "left"
        y_config.legend.show = True
        y_config.legend.loc = "below"
        y_config.legend.orientation = "horizontal"
        y_config.legend.bbox_to_anchor = (1, -0.4)

        # remove legend from class_kwargs to update legend args for y axis plot
        class_kwargs.pop("legend", None)
        class_kwargs.pop("xlabel", None)
        class_kwargs.pop("ylabel", None)

        color_gen = ColorGenerator()

        y_plot_obj = self.get_line_renderer(
            y_data, z, y, by=self.by, _config=y_config, **class_kwargs
        )
        y_fig = y_plot_obj.generate(line_color=color_gen)
        self.plot_x_axis_line(y_fig)
        self.ax_grid[1, 0].set_xlim((0, y_data[z].max() + y_data[z].max() * 0.1))
        self.ax_grid[1, 0].invert_xaxis()
        self.ax_grid[1, 0].set_title(None)
        self.ax_grid[1, 0].set_xlabel(self.zlabel)
        self.ax_grid[1, 0].set_ylabel(self.ylabel)
        self.ax_grid[1, 0].set_ylim(self.ax_grid[1, 1].get_ylim())

    def create_main_plot(self, x, y, z, class_kwargs, other_kwargs):
        scatterPlot = self.get_scatter_renderer(
            self.data, x, y, z=z, fig=self.fig, **class_kwargs
        )
        scatterPlot.generate(z=z, **other_kwargs)

        if self.annotation_data is not None:
            self._add_box_boundaries(self.annotation_data)

    def create_main_plot_marginals(self, x, y, z, class_kwargs, other_kwargs):
        scatterPlot = self.get_scatter_renderer(
            self.data, x, y, z=z, fig=self.ax_grid[1, 1], **class_kwargs
        )
        scatterPlot.generate(
            z=z,
            **other_kwargs,
        )
        self.ax_grid[1, 1].set_title(None)
        self.ax_grid[1, 1].set_xlabel(self.xlabel)
        self.ax_grid[1, 1].set_ylabel(None)
        self.ax_grid[1, 1].set_yticklabels([])
        self.ax_grid[1, 1].set_yticks([])
        self.ax_grid[1, 1].legend_ = None

    def _add_box_boundaries(self, annotation_data):
        if self.by is not None:
            legend = self.fig.get_legend()
            self.fig.add_artist(legend)

        color_gen = ColorGenerator(
            colormap=self.feature_config.colormap, n=annotation_data.shape[0]
        )
        legend_items = []

        for idx, (_, feature) in enumerate(annotation_data.iterrows()):
            x0 = feature["leftWidth"]
            x1 = feature["rightWidth"]
            y0 = feature["IM_leftWidth"]
            y1 = feature["IM_rightWidth"]

            # Calculate center points and dimensions
            width = abs(x1 - x0)
            height = abs(y1 - y0)

            color = next(color_gen)
            custom_lines = Rectangle(
                (x0, y0),
                width,
                height,
                fill=False,
                edgecolor=color,
                linestyle=self.feature_config.line_type,
                linewidth=self.feature_config.line_width,
            )
            self.fig.add_patch(custom_lines)

            if "name" in annotation_data.columns:
                use_name = feature["name"]
            else:
                use_name = f"Feature {idx}"
            if "q_value" in annotation_data.columns:
                legend_labels = f"{use_name} (q-value: {feature['q_value']:.4f})"
            else:
                legend_labels = f"{use_name}"

        # Add legend
        if self.feature_config.legend.show:
            matplotlibLegendLoc = LegendConfig._matplotlibLegendLocationMapper(
                self.feature_config.legend.loc
            )
            self.fig.legend(
                [custom_lines],
                [legend_labels],
                loc=matplotlibLegendLoc,
                title=self.feature_config.legend.title,
                prop={"size": self.feature_config.legend.fontsize},
                bbox_to_anchor=self.feature_config.legend.bbox_to_anchor,
            )

    # since matplotlib is not interactive cannot implement the following methods
    def get_manual_bounding_box_coords(self):
        pass



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Tuple, Literal, Union, List, Dict
import importlib
import types
import re

from pandas import cut
from pandas.core.frame import DataFrame
from pandas.core.dtypes.generic import ABCDataFrame
from pandas.core.dtypes.common import is_integer
from pandas.util._decorators import Appender

from numpy import log1p

from ._config import LegendConfig, FeatureConfig, _BasePlotConfig
from ._misc import ColorGenerator


_common_kinds = ("line", "vline", "scatter")
_msdata_kinds = ("chromatogram", "mobilogram", "spectrum", "peakmap")
_all_kinds = _common_kinds + _msdata_kinds
_entrypoint_backends = ("ms_matplotlib", "ms_bokeh", "ms_plotly")

_baseplot_doc = f"""
    Plot method for creating plots from a Pandas DataFrame.

    Parameters
    ----------
    data : pandas.DataFrame or numpy.ndarray
        The data to be plotted.
    x : str or None, optional
        The column name for the x-axis data.
    y : str or None, optional
        The column name for the y-axis data.
    z : str or None, optional
        The column name for the z-axis data (for 3D plots).
    kind : str, optional
        The kind of plot to create. One of: {_all_kinds}
    by : str or None, optional
        Column in the DataFrame to group by.
    relative_intensity : bool, default False
        Whether to use relative intensity for the y-axis.
    subplots : bool or None, optional
        Whether to create separate subplots for each column.
    sharex, sharey : bool or None, optional
        Whether to share x or y axes among subplots.
    height, width : int or None, optional
        The height and width of the figure in pixels.
    grid : bool or None, optional
        Whether to show the grid on the plot.
    toolbar_location : str or None, optional
        The location of the toolbar (e.g., 'above', 'below', 'left', 'right').
    fig : figure or None, optional
        An existing figure object to plot on.
    title : str or None, optional
        The title of the plot.
    xlabel, ylabel : str or None, optional
        Labels for the x and y axes.
    x_axis_location, y_axis_location : str or None, optional
        The location of the x and y axes (e.g., 'bottom', 'top', 'left', 'right').
    line_type : str or None, optional
        The type of line to use (e.g., 'solid', 'dashed', 'dotted').
    line_width : float or None, optional
        The width of the lines in the plot.
    min_border : int or None, optional
        The minimum border size around the plot.
    show_plot : bool or None, optional
        Whether to display the plot immediately after creation.
    legend : LegendConfig or dict or None, optional
        Configuration for the plot legend.
    feature_config : FeatureConfig or dict or None, optional
        Configuration for additional plot features.
    backend : str, default None
        Backend to use instead of the backend specified in the option
        ``plotting.backend``. For pyopenms_viz, options are one of {_entrypoint_backends} Alternatively, to
        specify the ``plotting.backend`` for the whole session, set
        ``pd.options.plotting.backend``.
    **kwargs
        Additional keyword arguments to be passed to the plotting function.

    Returns
    -------
    None

    Examples
    --------
    >>> import pandas as pd
    >>> 
    >>> data = pd.DataFrame(dict'x': [1, 2, 3], 'y': [4, 5, 6]))
    >>> data.plot(x='x', y='y', kind='spectrum', backend='pomsvim')
    """

APPEND_PLOT_DOC = Appender(_baseplot_doc)


class BasePlot(ABC):
    """
    This class shows functions which must be implemented by all backends
    """

    def __init__(
        self,
        data,
        x: str | None = None,
        y: str | None = None,
        z: str | None = None,
        kind=None,
        by: str | None = None,
        relative_intensity: bool = False,
        subplots: bool | None = None,
        sharex: bool | None = None,
        sharey: bool | None = None,
        height: int | None = None,
        width: int | None = None,
        grid: bool | None = None,
        toolbar_location: str | None = None,
        fig: "figure" | None = None,
        title: str | None = None,
        xlabel: str | None = None,
        ylabel: str | None = None,
        x_axis_location: str | None = None,
        y_axis_location: str | None = None,
        line_type: str | None = None,
        line_width: float | None = None,
        min_border: int | None = None,
        show_plot: bool | None = None,
        legend: LegendConfig | Dict | None = None,
        feature_config: FeatureConfig | Dict | None = None,
        _config: _BasePlotConfig | None = None,
        **kwargs,
    ) -> None:

        # Data attributes
        self.data = data.copy()
        self.kind = kind
        self.by = by
        self.relative_intensity = relative_intensity

        # Plotting attributes
        self.subplots = subplots
        self.sharex = sharex
        self.sharey = sharey
        self.height = height
        self.width = width
        self.grid = grid
        self.toolbar_location = toolbar_location
        self.fig = fig
        self.title = title
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.x_axis_location = x_axis_location
        self.y_axis_location = y_axis_location
        self.line_type = line_type
        self.line_width = line_width
        self.min_border = min_border
        self.show_plot = show_plot

        self.legend = legend
        self.feature_config = feature_config

        self._config = _config

        if _config is not None:
            self._update_from_config(_config)

        if self.legend is not None and isinstance(self.legend, dict):
            self.legend = LegendConfig.from_dict(self.legend)

        if self.feature_config is not None and isinstance(self.feature_config, dict):
            self.feature_config = FeatureConfig.from_dict(self.feature_config)

        ### get x and y data
        if self._kind in {
            "line",
            "vline",
            "scatter",
            "chromatogram",
            "mobilogram",
            "spectrum",
            "peakmap",
            "complex",
        }:
            self.x = self._verify_column(x, "x")
            self.y = self._verify_column(y, "y")

        if self._kind in {"peakmap"}:
            self.z = self._verify_column(z, "z")

        if self.by is not None:
            # Ensure by column data is string
            self.by = self._verify_column(by, "by")
            self.data[self.by] = self.data[self.by].astype(str)

        self._load_extension()
        self._create_figure()

    def _verify_column(self, colname: str | int, name: str) -> str:
        """fetch data from column name

        Args:
            colname (str | int): column name of data to fetch or the index of the column to fetch
            name (str): name of the column e.g. x, y, z for error message

        Returns:
            pd.Series: pandas series or None

        Raises:
            ValueError: if colname is None
            KeyError: if colname is not in data
            ValueError: if colname is not numeric
        """

        def holds_integer(column) -> bool:
            return column.inferred_type in {"integer", "mixed-integer"}

        if colname is None:
            raise ValueError(f"For `{self.kind}` plot, `{name}` must be set")

        # if integer is supplied get the corresponding column associated with that index
        if is_integer(colname) and not holds_integer(self.data.columns):
            if colname >= len(self.data.columns):
                print(self.data.columns)
                raise ValueError(
                    f"Column index `{colname}` out of range, `{name}` could not be set"
                )
            else:
                colname = self.data.columns[colname]
        else:  # assume column name is supplied
            if colname not in self.data.columns:
                raise KeyError(
                    f"Column `{colname}` not in data, `{name}` could not be set"
                )

        # checks passed return column name
        return colname

    @property
    @abstractmethod
    def _kind(self) -> str:
        """
        The kind of plot to assemble. Must be overridden by subclasses.
        """
        raise NotImplementedError

    @property
    def _interactive(self) -> bool:
        """
        Whether the plot is interactive. Must be overridden by subclasses
        """
        return NotImplementedError

    def _update_from_config(self, config) -> None:
        """
        Updates the plot configuration based on the provided `config` object.

        Args:
            config (Config): The configuration object containing the plot settings.

        Returns:
            None
        """
        for attr, value in config.__dict__.items():
            if (
                value is not None
                and hasattr(self, attr)
                and self.__dict__[attr] is None
            ):
                setattr(self, attr, value)

    def _separate_class_kwargs(self, **kwargs):
        """
        Separates the keyword arguments into class-specific arguments and other arguments.

        Parameters:
            **kwargs: Keyword arguments passed to the method.

        Returns:
            class_kwargs: A dictionary containing the class-specific keyword arguments.
            other_kwargs: A dictionary containing the remaining keyword arguments.

        """
        class_kwargs = {k: v for k, v in kwargs.items() if k in dir(self)}
        other_kwargs = {k: v for k, v in kwargs.items() if k not in dir(self)}
        return class_kwargs, other_kwargs

    @abstractmethod
    def _load_extension(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def _create_figure(self) -> None:
        raise NotImplementedError

    def _make_plot(self, fig, **kwargs) -> None:
        # Check for tooltips in kwargs and pop
        tooltips = kwargs.pop("tooltips", None)
        custom_hover_data = kwargs.pop("custom_hover_data", None)

        newlines, legend = self.plot(fig, self.data, self.x, self.y, self.by, **kwargs)

        if legend is not None:
            self._add_legend(newlines, legend)
        self._update_plot_aes(newlines, **kwargs)

        if tooltips is not None and self._interactive:
            self._add_tooltips(newlines, tooltips, custom_hover_data)

    @abstractmethod
    def plot(cls, fig, data, x, y, by: str | None = None, **kwargs):
        """
        Create the plot
        """
        pass

    @abstractmethod
    def _update_plot_aes(self, fig, **kwargs):
        pass

    @abstractmethod
    def _add_legend(self, fig, legend):
        pass

    @abstractmethod
    def _modify_x_range(
        self, x_range: Tuple[float, float], padding: Tuple[float, float] | None = None
    ):
        """
        Modify the x-axis range.

        Args:
            x_range (Tuple[float, float]): The desired x-axis range.
            padding (Tuple[float, float] | None, optional): The padding to be applied to the x-axis range, in decimal percent. Defaults to None.
        """
        pass

    @abstractmethod
    def _modify_y_range(
        self, y_range: Tuple[float, float], padding: Tuple[float, float] | None = None
    ):
        """
        Modify the y-axis range.

        Args:
            y_range (Tuple[float, float]): The desired y-axis range.
            padding (Tuple[float, float] | None, optional): The padding to be applied to the x-axis range, in decimal percent. Defaults to None.
        """
        pass

    def generate(self, **kwargs):
        """
        Generate the plot
        """
        self._make_plot(self.fig, **kwargs)
        return self.fig

    @abstractmethod
    def show(self):
        pass

    # methods only for interactive plotting
    @abstractmethod
    def _add_tooltips(self, fig, tooltips):
        pass

    @abstractmethod
    def _add_bounding_box_drawer(self, fig, **kwargs):
        pass

    def _add_bounding_vertical_drawer(self, fig, **kwargs):
        pass


class LinePlot(BasePlot, ABC):
    @property
    def _kind(self):
        return "line"


class VLinePlot(BasePlot, ABC):
    @property
    def _kind(self):
        return "vline"

    def _add_annotations(
        self,
        fig,
        ann_texts: list[str],
        ann_xs: list[float],
        ann_ys: list[float],
        ann_colors: list[str],
    ):
        """
        Add annotations to a VLinePlot figure.

        Parameters:
        fig: The figure to add annotations to.
        ann_texts (list[str]): List of texts for the annotations.
        ann_xs (list[float]): List of x-coordinates for the annotations.
        ann_ys (list[float]): List of y-coordinates for the annotations.
        ann_colors: (list[str]): List of colors for annotation text.
        """
        pass

class ScatterPlot(BasePlot, ABC):
    @property
    def _kind(self):
        return "scatter"


class BaseMSPlot(BasePlot, ABC):
    """
    Abstract class for complex plots, such as chromatograms and mobilograms which are made up of simple plots such as ScatterPlots, VLines and LinePlots.

    Args:
        BasePlot (_type_): _description_
        ABC (_type_): _description_
    """

    @abstractmethod
    def get_line_renderer(self, data, x, y, **kwargs):
        pass

    @abstractmethod
    def get_vline_renderer(self, data, x, y, **kwargs):
        pass

    @abstractmethod
    def get_scatter_renderer(self, data, x, y, **kwargs):
        pass

    @abstractmethod
    def plot_x_axis_line(self, fig):
        """
        plot line across x axis
        """
        pass

    @abstractmethod
    def _create_tooltips(self, entries: dict, index: bool = True):
        """
        Create tooltipis based on entries dictionary with keys: label for tooltip and values: column names.

        entries = {
            "m/z": "mz"
            "Retention Time: "RT"
        }

        Will result in tooltips where label and value are separated by colon:

        m/z: 100.5
        Retention Time: 50.1

        Parameters:
            entries (dict): Which data to put in tool tip and how display it with labels as keys and column names as values.
            index (bool, optional): Wether to show dataframe index in tooltip. Defaults to True.

        Returns:
            Tooltip text.
            Tooltip data.
        """
        pass


class ChromatogramPlot(BaseMSPlot, ABC):
    @property
    def _kind(self):
        return "chromatogram"

    def __init__(
        self, data, x, y, annotation_data: DataFrame | None = None, **kwargs
    ) -> None:

        # Set default config attributes if not passed as keyword arguments
        kwargs["_config"] = _BasePlotConfig(kind=self._kind)

        super().__init__(data, x, y, **kwargs)

        if annotation_data is not None:
            self.annotation_data = annotation_data.copy()
        else:
            self.annotation_data = None
        self.label_suffix = self.x  # set label suffix for bounding box

        self.plot(self.data, self.x, self.y, **kwargs)
        if self.show_plot:
            self.show()

    def plot(self, data, x, y, **kwargs):
        """
        Create the plot
        """
        color_gen = ColorGenerator()
        tooltip_entries = {"retention time": x, "intensity": y}
        if "Annotation" in self.data.columns:
            tooltip_entries["annotation"] = "Annotation"
        if "product_mz" in self.data.columns:
            tooltip_entries["product m/z"] = "product_mz"
        TOOLTIPS, custom_hover_data = self._create_tooltips(tooltip_entries)
        kwargs.pop(
            "fig", None
        )  # remove figure from **kwargs if exists, use the ChromatogramPlot figure object instead of creating a new figure
        linePlot = self.get_line_renderer(data, x, y, fig=self.fig, **kwargs)
        self.fig = linePlot.generate(
            line_color=color_gen, tooltips=TOOLTIPS, custom_hover_data=custom_hover_data
        )

        self._modify_y_range((0, self.data[y].max()), (0, 0.1))

        self.manual_boundary_renderer = (
            self._add_bounding_vertical_drawer(self.fig) if self._interactive else None
        )

        if self.annotation_data is not None:
            self._add_peak_boundaries(self.annotation_data)

    @abstractmethod
    def _add_peak_boundaries(self, annotation_data):
        """
        Prepare data for adding peak boundaries to the plot.

        Args:
            annotation_data (DataFrame): The feature data containing the peak boundaries.

        Returns:
            None
        """
        pass


class MobilogramPlot(ChromatogramPlot, ABC):

    @property
    def _kind(self):
        return "mobilogram"

    def __init__(
        self, data, x, y, annotation_data: DataFrame | None = None, **kwargs
    ) -> None:
        super().__init__(data, x, y, annotation_data=annotation_data, **kwargs)

    def plot(self, data, x, y, **kwargs):
        super().plot(data, x, y, **kwargs)
        self._modify_y_range((0, self.data[y].max()), (0, 0.1))


class SpectrumPlot(BaseMSPlot, ABC):
    @property
    def _kind(self):
        return "spectrum"

    def __init__(
        self,
        data: DataFrame,
        x: str,
        y: str,
        reference_spectrum: DataFrame | None = None,
        mirror_spectrum: bool = False,
        relative_intensity: bool = False,
        peak_color: str | None = None,
        annotate_top_n_peaks: int | None | Literal["all"] = 5,
        annotate_mz: bool = True,
        ion_annotation: str | None = None,
        sequence_annotation: str | None = None,
        custom_annotation: str | None = None,
        annotation_color: str | None = None,
        **kwargs,
    ) -> None:

        # Set default config attributes if not passed as keyword arguments
        kwargs["_config"] = _BasePlotConfig(kind=self._kind)

        super().__init__(data, x, y, **kwargs)

        self.reference_spectrum = reference_spectrum
        self.mirror_spectrum = mirror_spectrum
        self.relative_intensity = relative_intensity
        self.peak_color = peak_color
        self.annotate_top_n_peaks = annotate_top_n_peaks
        self.annotate_mz = annotate_mz
        self.ion_annotation = ion_annotation
        self.sequence_annotation = sequence_annotation
        self.custom_annotation = custom_annotation
        self.annotation_color = annotation_color

        self.plot(x, y, **kwargs)
        # Show plot
        if self.show_plot:
            self.show()

    def plot(self, x, y, **kwargs):
        """Standard spectrum plot with m/z on x-axis, intensity on y-axis and optional mirror spectrum."""
        # Prepare data
        spectrum, reference_spectrum = self._prepare_data(
            self.data, y, self.reference_spectrum
        )
        kwargs.pop("fig", None)  # remove figure from **kwargs if exists

        entries = {"m/z": x, "intensity": y}
        for optional in (
            "native_id",
            self.ion_annotation,
            self.sequence_annotation,
        ):
            if optional in self.data.columns:
                entries[optional.replace("_", " ")] = optional

        tooltips, custom_hover_data = self._create_tooltips(
            entries=entries, index=False
        )

        spectrumPlot = self.get_vline_renderer(spectrum, x, y, fig=self.fig, **kwargs)

        color_gen = self._get_colors(spectrum, "peak")

        self.fig = spectrumPlot.generate(
            line_color=color_gen, tooltips=tooltips, custom_hover_data=custom_hover_data
        )

        # Annotations for spectrum
        ann_texts, ann_xs, ann_ys, ann_colors = self._get_annotations(spectrum, x, y)
        spectrumPlot._add_annotations(self.fig, ann_texts, ann_xs, ann_ys, ann_colors)

        # Mirror spectrum
        if self.mirror_spectrum and reference_spectrum is not None:
            # Set intensity to negative values
            reference_spectrum[y] = reference_spectrum[y] * -1

            mirror_spectrum = self.get_vline_renderer(
                reference_spectrum, x, y, fig=self.fig, **kwargs
            )

            color_gen = self._get_colors(reference_spectrum, "peak")

            mirror_spectrum.generate(line_color=color_gen)
            self.plot_x_axis_line(self.fig)

            # Annotations for reference spectrum
            ann_texts, ann_xs, ann_ys, ann_colors = self._get_annotations(
                reference_spectrum, x, y
            )
            spectrumPlot._add_annotations(
                self.fig, ann_texts, ann_xs, ann_ys, ann_colors
            )

        # Adjust x axis padding (Plotly cuts outermost peaks)
        min_values = [spectrum[x].min()]
        max_values = [spectrum[x].max()]
        if reference_spectrum is not None:
            min_values.append(reference_spectrum[x].min())
            max_values.append(reference_spectrum[x].max())
        self._modify_x_range((min(min_values), max(max_values)), padding=(0.20, 0.20))

        # Adjust y axis padding (annotations should stay inside plot)
        max_value = spectrum[y].max()
        min_value = 0
        min_padding = 0
        max_padding = 0.15
        if reference_spectrum is not None and self.mirror_spectrum:
            min_value = reference_spectrum[y].min()
            min_padding = -0.2
            max_padding = 0.4

        self._modify_y_range((min_value, max_value), padding=(min_padding, max_padding))

    def _prepare_data(
        self,
        spectrum: DataFrame,
        y: str,
        reference_spectrum: Union[DataFrame, None],
    ) -> tuple[list, list]:
        """Prepares data for plotting based on configuration (copy, relative intensity)."""

        # copy spectrum data to not modify the original
        spectrum = spectrum.copy()
        reference_spectrum = (
            self.reference_spectrum.copy() if reference_spectrum is not None else None
        )

        # Convert to relative intensity if required
        if self.relative_intensity or self.mirror_spectrum:
            spectrum[y] = spectrum[y] / spectrum[y].max() * 100
            if reference_spectrum is not None:
                reference_spectrum[y] = (
                    reference_spectrum[y] / reference_spectrum[y].max() * 100
                )
        return spectrum, reference_spectrum

    def _get_colors(
        self, data: DataFrame, kind: Literal["peak", "annotation"] | None = None
    ):
        """Get color generators for peaks or annotations based on config."""
        # Top priority: custom color
        if kind is not None:
            if kind == "peak" and self.peak_color in data.columns:
                return ColorGenerator(data[self.peak_color])
            elif kind == "annotation" and self.annotation_color in data.columns:
                return ColorGenerator(data[self.annotation_color])
        # Colors based on ion annotation for peaks and annotation text
        if self.ion_annotation is not None and self.ion_annotation in data.columns:
            return self._get_ion_color_annotation(data)
        # Color peaks of a group with the same color (from default colors)
        if self.by:
            if self.by in data.columns:
                uniques = data[self.by].unique()
                color_gen = ColorGenerator()
                colors = [next(color_gen) for _ in range(len(uniques))]
                color_map = {uniques[i]: colors[i] for i in range(len(colors))}
                all_colors = data[self.by].apply(lambda x: color_map[x])
                return ColorGenerator(all_colors)
        # Lowest priority: return the first default color
        return ColorGenerator(None, 1)

    def _get_annotations(self, data: DataFrame, x: str, y: str):
        """Create annotations for each peak. Return lists of texts, x and y locations and colors."""
        color_gen = self._get_colors(data, "annotation")

        data["color"] = [next(color_gen) for _ in range(len(data))]

        ann_texts = []
        top_n = self.annotate_top_n_peaks
        if top_n == "all":
            top_n = len(data)
        elif top_n is None:
            top_n = 0
        # sort values for top intensity peaks on top (ascending for reference spectra with negative values)
        data = data.sort_values(
            y, ascending=True if data[y].min() < 0 else False
        ).reset_index()
        for i, row in data.iterrows():
            texts = []
            if i < top_n:
                if self.annotate_mz:
                    texts.append(str(round(row[x], 4)))
                if self.ion_annotation and self.ion_annotation in data.columns:
                    texts.append(str(row[self.ion_annotation]))
                if (
                    self.sequence_annotation
                    and self.sequence_annotation in data.columns
                ):
                    texts.append(str(row[self.sequence_annotation]))
                if self.custom_annotation and self.custom_annotation in data.columns:
                    texts.append(str(row[self.custom_annotation]))
            ann_texts.append("\n".join(texts))
        return ann_texts, data[x].tolist(), data[y].tolist(), data["color"].tolist()

    def _get_ion_color_annotation(self, data: DataFrame) -> str:
        """Retrieve the color associated with a specific ion annotation from a predefined colormap."""
        colormap = {
            "a": ColorGenerator.color_blind_friendly_map[ColorGenerator.Colors.PURPLE],
            "b": ColorGenerator.color_blind_friendly_map[ColorGenerator.Colors.BLUE],
            "c": ColorGenerator.color_blind_friendly_map[
                ColorGenerator.Colors.LIGHTBLUE
            ],
            "x": ColorGenerator.color_blind_friendly_map[ColorGenerator.Colors.YELLOW],
            "y": ColorGenerator.color_blind_friendly_map[ColorGenerator.Colors.RED],
            "z": ColorGenerator.color_blind_friendly_map[ColorGenerator.Colors.ORANGE],
        }

        def get_ion_color(ion):
            if isinstance(ion, str):
                for key in colormap.keys():
                    # Exact matches
                    if ion == key:
                        return colormap[key]
                    # Fragment ions via regex
                    x = re.search(r"^[abcxyz]{1}[0-9]*[+-]$", ion)
                    if x:
                        return colormap[ion[0]]
            return ColorGenerator.color_blind_friendly_map[
                ColorGenerator.Colors.DARKGRAY
            ]

        colors = data[self.ion_annotation].apply(get_ion_color)
        return ColorGenerator(colors)


class PeakMapPlot(BaseMSPlot, ABC):
    # need to inherit from ChromatogramPlot and SpectrumPlot for get_line_renderer and get_vline_renderer methods respectively
    @property
    def _kind(self):
        return "peakmap"

    def __init__(
        self,
        data,
        x,
        y,
        z,
        zlabel=None,
        add_marginals=False,
        annotation_data: DataFrame | None = None,
        bin_peaks: Union[Literal["auto"], bool] = "auto",
        num_x_bins: int = 50,
        num_y_bins: int = 50,
        z_log_scale: bool = False,
        **kwargs,
    ) -> None:
        # Copy data since it will be modified
        data = data.copy()

        # Set default config attributes if not passed as keyword arguments
        kwargs["_config"] = _BasePlotConfig(kind=self._kind)

        if add_marginals:
            kwargs["_config"].title = None

        self.zlabel = zlabel
        self.add_marginals = add_marginals

        if annotation_data is not None:
            self.annotation_data = annotation_data.copy()
        else:
            self.annotation_data = None

        # Convert intensity values to relative intensity if required
        relative_intensity = kwargs.pop("relative_intensity", False)
        if relative_intensity:
            data[z] = data[z] / max(data[z]) * 100

        # Bin peaks if required
        if bin_peaks == True or (
            data.shape[0] > num_x_bins * num_y_bins
            and bin_peaks == "auto"
        ):
            data[x] = cut(data[x], bins=num_x_bins)
            data[y] = cut(data[y], bins=num_y_bins)

            # Group by x and y bins and calculate the mean intensity within each bin
            data = (
                data.groupby([x, y], observed=True)
                .agg({z: "mean"})
                .reset_index()
            )
            data[x] = data[x].apply(lambda interval: interval.mid).astype(float)
            data[y] = data[y].apply(lambda interval: interval.mid).astype(float)
            data = data.fillna(0)

        # Log intensity scale
        if z_log_scale:
            data[z] = log1p(data[z])

        # Sort values by intensity in ascending order to plot highest intensity peaks last
        data = data.sort_values(z)

        super().__init__(data, x, y, z=z, **kwargs)

        self.plot(x, y, z, **kwargs)
        if self.show_plot:
            self.show()

    def plot(self, x, y, z, **kwargs):
        class_kwargs, other_kwargs = self._separate_class_kwargs(**kwargs)

        if self.add_marginals:
            self.create_main_plot_marginals(x, y, z, class_kwargs, other_kwargs)
        else:
            self.create_main_plot(x, y, z, class_kwargs, other_kwargs)

        self.manual_bbox_renderer = (
            self._add_bounding_box_drawer(self.fig) if self._interactive else None
        )

        if self.add_marginals:
            # remove 'config' from class_kwargs
            class_kwargs_copy = class_kwargs.copy()
            class_kwargs_copy.pop("_config", None)
            class_kwargs_copy.pop("by", None)

            x_fig = self.create_x_axis_plot(x, z, class_kwargs_copy)

            y_fig = self.create_y_axis_plot(y, z, class_kwargs_copy)

            self.combine_plots(x_fig, y_fig)

    @staticmethod
    def _integrate_data_along_dim(
        data: DataFrame, group_cols: List[str] | str, integrate_col: str
    ) -> DataFrame:
        # First fill NaNs with 0s for numerical columns and '.' for categorical columns
        grouped = (
            data.apply(
                lambda x: x.fillna(0) if x.dtype.kind in "biufc" else x.fillna(".")
            )
            .groupby(group_cols)[integrate_col]
            .sum()
            .reset_index()
        )
        return grouped

    @abstractmethod
    def create_main_plot(self, x, y, z, class_kwargs, other_kwargs):
        pass

    # by default the main plot with marginals is plotted the same way as the main plot unless otherwise specified
    def create_main_plot_marginals(self, x, y, z, class_kwargs, other_kwargs):
        self.create_main_plot(x, y, z, class_kwargs, other_kwargs)

    @abstractmethod
    def create_x_axis_plot(self, x, z, class_kwargs) -> "figure":
        # get cols to integrate over and exclude y and z
        group_cols = [x]
        if self.by is not None:
            group_cols.append(self.by)

        x_data = self._integrate_data_along_dim(self.data, group_cols, z)

        x_config = self._config.copy()
        x_config.ylabel = self.zlabel
        x_config.y_axis_location = "right"
        x_config.legend.show = True
        x_config.legend.loc = "right"

        color_gen = ColorGenerator()

        # remove legend from class_kwargs to update legend args for x axis plot
        class_kwargs.pop("legend", None)
        class_kwargs.pop("ylabel", None)

        x_plot_obj = self.get_line_renderer(
            x_data, x, z, by=self.by, _config=x_config, **class_kwargs
        )
        x_fig = x_plot_obj.generate(line_color=color_gen)
        self.plot_x_axis_line(x_fig)

        return x_fig

    @abstractmethod
    def create_y_axis_plot(self, y, z, class_kwargs) -> "figure":
        group_cols = [y]
        if self.by is not None:
            group_cols.append(self.by)

        y_data = self._integrate_data_along_dim(self.data, group_cols, z)

        y_config = self._config.copy()
        y_config.xlabel = self.zlabel
        y_config.ylabel = self.ylabel
        y_config.y_axis_location = "left"
        y_config.legend.show = True
        y_config.legend.loc = "below"

        # remove legend from class_kwargs to update legend args for y axis plot
        class_kwargs.pop("legend", None)
        class_kwargs.pop("xlabel", None)

        color_gen = ColorGenerator()

        y_plot_obj = self.get_line_renderer(
            y_data, z, y, by=self.by, _config=y_config, **class_kwargs
        )
        y_fig = y_plot_obj.generate(line_color=color_gen)
        self.plot_x_axis_line(y_fig)

        return y_fig

    @abstractmethod
    def combine_plots(self, x_fig, y_fig):
        pass

    @abstractmethod
    def _add_box_boundaries(self, annotation_data):
        """
        Prepare data for adding box boundaries to the plot.

        Args:
            annotation_data (DataFrame): The feature data containing the box boundaries.

        Returns:
            None
        """
        pass


class PlotAccessor:
    """
    Make plots of MassSpec data using dataframes

    """

    _common_kinds = ("line", "vline", "scatter")
    _msdata_kinds = ("chromatogram", "mobilogram", "spectrum", "peakmap")
    _all_kinds = _common_kinds + _msdata_kinds

    def __init__(self, data: DataFrame) -> None:
        self._parent = data

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        backend_name = kwargs.get("backend", None)
        if backend_name is None:
            backend_name = "matplotlib"

        plot_backend = _get_plot_backend(backend_name)

        x, y, kind, kwargs = self._get_call_args(
            plot_backend.__name__, self._parent, args, kwargs
        )

        if kind not in self._all_kinds:
            raise ValueError(
                f"{kind} is not a valid plot kind "
                f"Valid plot kinds: {self._all_kinds}"
            )

        # Call the plot method of the selected backend
        if "backend" in kwargs:
            kwargs.pop("backend")
        return plot_backend.plot(self._parent, x=x, y=y, kind=kind, **kwargs)

    @staticmethod
    def _get_call_args(backend_name: str, data: DataFrame, args, kwargs):
        """
        Get the arguments to pass to the plotting backend.

        Parameters
        ----------
        backend_name : str
            The name of the backend.
        data : DataFrame
            The data to plot.
        args : tuple
            The positional arguments passed to the plotting function.
        kwargs : dict
            The keyword arguments passed to the plotting function.

        Returns
        -------
        dict
            The arguments to pass to the plotting backend.
        """
        if isinstance(data, ABCDataFrame):
            arg_def = [
                ("x", None),
                ("y", None),
                ("kind", "line"),
                ("by", None),
                ("subplots", None),
                ("sharex", None),
                ("sharey", None),
                ("height", None),
                ("width", None),
                ("grid", None),
                ("toolbar_location", None),
                ("fig", None),
                ("title", None),
                ("xlabel", None),
                ("ylabel", None),
                ("x_axis_location", None),
                ("y_axis_location", None),
                ("line_type", None),
                ("line_width", None),
                ("min_border", None),
                ("show_plot", None),
                ("legend", None),
                ("feature_config", None),
                ("_config", None),
                ("backend", backend_name),
            ]
        else:
            raise ValueError(
                f"Unsupported data type: {type(data).__name__}, expected DataFrame."
            )

        pos_args = {name: value for (name, _), value in zip(arg_def, args)}

        kwargs = dict(arg_def, **pos_args, **kwargs)

        x = kwargs.pop("x", None)
        y = kwargs.pop("y", None)
        kind = kwargs.pop("kind", "line")
        return x, y, kind, kwargs


_backends: dict[str, types.ModuleType] = {}


def _load_backend(backend: str) -> types.ModuleType:
    """
    Load a plotting backend.

    Parameters
    ----------
    backend : str
        The identifier for the backend. Either "bokeh", "matplotlib", "plotly",
        or a module name.

    Returns
    -------
    types.ModuleType
        The imported backend.
    """
    if backend == "bokeh":
        try:
            module = importlib.import_module("pyopenms_viz.plotting._bokeh")
        except ImportError:
            raise ImportError(
                "Bokeh is required for plotting when the 'bokeh' backend is selected."
            ) from None
        return module

    elif backend == "matplotlib":
        try:
            module = importlib.import_module("pyopenms_viz.plotting._matplotlib")
        except ImportError:
            raise ImportError(
                "Matplotlib is required for plotting when the 'matplotlib' backend is selected."
            ) from None
        return module

    elif backend == "plotly":
        try:
            module = importlib.import_module("pyopenms_viz.plotting._plotly")
        except ImportError:
            raise ImportError(
                "Plotly is required for plotting when the 'plotly' backend is selected."
            ) from None
        return module

    raise ValueError(
        f"Could not find plotting backend '{backend}'. Needs to be one of 'bokeh', 'matplotlib', or 'plotly'."
    )


def _get_plot_backend(backend: str | None = None):

    backend_str: str = backend or "matplotlib"

    if backend_str in _backends:
        return _backends[backend_str]

    module = _load_backend(backend_str)
    _backends[backend_str] = module
    return module



from enum import Enum, auto
import matplotlib.pyplot as plt
import numpy as np
from itertools import cycle
from typing import Literal


class ColorGenerator:
    """
    A class that generates colors for plotting.

    Attributes:
        color_blind_friendly_map (dict): A dictionary that maps the enum values to their corresponding colors.

    Methods:
        __init__(self, colormap=None, n=None): Initializes a ColorGenerator object.
        __iter__(self): Returns the ColorGenerator object itself.
        __next__(self): Returns the next color in the color cycle.
        generate_colors(self, n=None): Generates a list of colors.
        _get_n_grayscale_colors(self, n: int) -> list: Returns n evenly spaced grayscale colors in hex format.
    """

    class Colors(Enum):
        """
        Enum class that defines color options.
        """

        BLUE = auto()
        RED = auto()
        LIGHTBLUE = auto()
        ORANGE = auto()
        PURPLE = auto()
        YELLOW = auto()
        DARKGRAY = auto()
        LIGHTGRAY = auto()

    color_blind_friendly_map = {
        Colors.BLUE: "#4575B4",
        Colors.RED: "#D73027",
        Colors.LIGHTBLUE: "#91BFDB",
        Colors.ORANGE: "#FC8D59",
        Colors.PURPLE: "#7B2C65",
        Colors.YELLOW: "#FCCF53",
        Colors.DARKGRAY: "#555555",
        Colors.LIGHTGRAY: "#BBBBBB",
    }

    def __init__(self, colormap=None, n=None):
        """
        Initializes a ColorGenerator object.

        Args:
            colormap (str or list, optional): The colormap to use for generating colors. Defaults to None.
            n (int, optional): The number of colors to generate. Defaults to None.
        """
        if colormap is None:
            colors = list(self.color_blind_friendly_map.values())
        else:
            if isinstance(colormap, str):
                if colormap.lower() == "grayscale":
                    colors = self._get_n_grayscale_colors(n)
                else:
                    cmap = plt.get_cmap(colormap, n)
                    colors = cmap(np.linspace(0, 1, n))
                    colors = [
                        "#{:02X}{:02X}{:02X}".format(
                            int(r * 255), int(g * 255), int(b * 255)
                        )
                        for r, g, b, _ in colors
                    ]
            else:
                colors = colormap
        if n is not None:
            colors = colors[:n]
        self.color_cycle = cycle(colors)

    def __iter__(self):
        """
        Returns the ColorGenerator object itself.
        """
        return self

    def __next__(self):
        """
        Returns the next color in the color cycle.
        """
        return next(self.color_cycle)

    def generate_colors(self, n=None):
        """
        Generates a list of colors.

        Args:
            n (int, optional): The number of colors to generate. Defaults to None.

        Returns:
            list: A list of colors.
        """
        if n is None:
            return next(self)
        else:
            return [next(self) for _ in range(n)]

    def _get_n_grayscale_colors(self, n: int) -> list:
        """
        Returns n evenly spaced grayscale colors in hex format.

        Args:
            n (int): The number of colors to generate.

        Returns:
            list: A list of grayscale colors in hex format.
        """
        hex_list = []
        for v in np.linspace(50, 200, n):
            hex_color = "#"
            for _ in range(3):
                hex_color += f"{int(round(v)):02x}"
            hex_list.append(hex_color)
        return hex_list


class MarkerShapeGenerator:
    """
    A class that generates colors for plotting.

    Attributes:
        color_blind_friendly_map (dict): A dictionary that maps the enum values to their corresponding colors.

    Methods:
        __init__(self, colormap=None, n=None): Initializes a ColorGenerator object.
        __iter__(self): Returns the ColorGenerator object itself.
        __next__(self): Returns the next color in the color cycle.
        generate_colors(self, n=None): Generates a list of colors.
    """

    class Shapes(Enum):
        """
        Enum class that defines marker shape options.
        """

        SQUARE = auto()
        CIRCLE = auto()
        DIAMOND = auto()
        CROSS = auto()
        X = auto()
        TRIANGLE = auto()
        DOWN = auto()
        LEFT = auto()
        RIGHT = auto()
        PENTAGON = auto()
        HEXAGON = auto()

    matplotlib_map = {
        Shapes.SQUARE: "s",
        Shapes.CIRCLE: "o",
        Shapes.DIAMOND: "d",
        Shapes.CROSS: "+",
        Shapes.X: "x",
        Shapes.TRIANGLE: "^",
        Shapes.DOWN: "v",
        Shapes.HEXAGON: "h",
    }

    plotly_map = {
        Shapes.SQUARE: "square",
        Shapes.CIRCLE: "circle",
        Shapes.DIAMOND: "diamond",
        Shapes.CROSS: "cross",
        Shapes.X: "x",
        Shapes.TRIANGLE: "triangle-up",
        Shapes.DOWN: "triangle-down",
        Shapes.HEXAGON: "hexagon",
    }

    bokeh_map = {
        Shapes.SQUARE: "square",
        Shapes.CIRCLE: "circle",
        Shapes.DIAMOND: "diamond",
        Shapes.CROSS: "cross",
        Shapes.X: "x",
        Shapes.TRIANGLE: "triangle",
        Shapes.DOWN: "inverted_triangle",
        Shapes.HEXAGON: "hex",
    }

    def __init__(
        self,
        shapes: list | None = None,
        engine: Literal["MATPLOTLIB", "PLOTLY", "BOKEH"] | None = None,
        n: int | None = None,
    ):
        """
        Initializes a MarkerShapeGenerator object.

        Args:
            shapes (list or None, optional): A pre-defined sequence of shapes. Defaults to None.
            engine (Literal["MATPLOTLIB", "PLOTLY", "BOKEH"] or None, optional): Plotting engine. Defaults to None.
            n (int or None, optional): The number of shapes to generate. Defaults to None.
        """
        if engine is None and shapes is None:
            raise ValueError("Pass either a list of shapes or an engine to initialize the MarkerShapeGenerator.")
        if shapes is not None:
            self.shape_cycle = cycle(shapes)
            return
        if engine == "MATPLOTLIB":
            shapes = list(self.matplotlib_map.values())
        elif engine == "PLOTLY":
            shapes = list(self.plotly_map.values())
        else:
            shapes = list(self.bokeh_map.values())
        if n is not None:
            shapes = shapes[:n]
        self.shape_cycle = cycle(shapes)
        return

    def __iter__(self):
        """
        Returns the MarkerShapeGenerator object itself.
        """
        return self

    def __next__(self):
        """
        Returns the next shape in the shape cycle.
        """
        return next(self.shape_cycle)
